# SPDX-License-Identifier: MIT
# Copyright © 2023 André Santos

###############################################################################
# Imports
###############################################################################

from hplrv.monitors import MonitorState

from .common_data import *

###############################################################################
# Global Scope No Timeout
###############################################################################

def globally_forbids():
    text = 'globally: a {x > 0} forbids b {x > 0}'
    traces = []
    # valid
    traces.append([])
    traces.append([ TimerEvent() ])
    traces.append([ SpamEvent('a', Point2D()) ])
    traces.append([ SpamEvent('b', Point2D(x=1)) ])
    traces.append([ TriggerEvent(Point2D(x=1), MonitorState.ACTIVE) ])
    traces.append([
        SpamEvent('b', Point2D(x=1)),
        TimerEvent(),
        SpamEvent('b', Point2D(x=2)),
    ])
    # invalid
    traces.append([
        TriggerEvent(Point2D(x=1), MonitorState.ACTIVE),
        SpamEvent('b', Point2D(x=0)),
        BehaviourEvent(Point2D(x=1), MonitorState.FALSE),
        SpamEvent('a', Point2D(x=2)),
        SpamEvent('b', Point2D(x=2)),
    ])
    return (text, traces)

def globally_forbids_ref():
    text = 'globally: a as A {x > 0} forbids b {x > 0 and x > @A.x}'
    traces = []
    # valid
    traces.append([])
    traces.append([ TimerEvent() ])
    traces.append([ SpamEvent('a', Point2D()) ])
    traces.append([ SpamEvent('b', Point2D(x=1)) ])
    traces.append([ TriggerEvent(Point2D(x=1), MonitorState.ACTIVE) ])
    traces.append([
        SpamEvent('b', Point2D(x=1)),
        TimerEvent(),
        SpamEvent('b', Point2D(x=2)),
    ])
    traces.append([
        TriggerEvent(Point2D(x=1), MonitorState.ACTIVE),
        SpamEvent('b', Point2D(x=1)),
        SpamEvent('b', Point2D(x=1)),
    ])
    # invalid
    traces.append([
        TriggerEvent(Point2D(x=1), MonitorState.ACTIVE),
        BehaviourEvent(Point2D(x=2), MonitorState.FALSE),
        SpamEvent('b', Point2D(x=2)),
    ])
    traces.append([
        TriggerEvent(Point2D(x=1), MonitorState.ACTIVE),
        TriggerEvent(Point2D(x=2), None),
        SpamEvent('b', Point2D(x=1)),
        BehaviourEvent(Point2D(x=2), MonitorState.FALSE),
        SpamEvent('a', Point2D(x=3)),
        SpamEvent('b', Point2D(x=4)),
    ])
    return (text, traces)

###############################################################################
# Global Scope With Timeout
###############################################################################

def globally_forbids_within():
    text = 'globally: a {x > 0} forbids b {x > 0} within 3 s'
    traces = []
    # valid
    traces.append([])
    traces.append([ TimerEvent() ])
    traces.append([ SpamEvent('a', Point2D()) ])
    traces.append([ SpamEvent('b', Point2D()) ])
    traces.append([
        SpamEvent('b', Point2D(x=1)),
        TimerEvent(),
        SpamEvent('b', Point2D(x=2)),
    ])
    traces.append([
        TriggerEvent(Point2D(x=1), MonitorState.ACTIVE),
        TimerEvent(),
        TimerEvent(),
        TimerEvent(state=MonitorState.SAFE),
        SpamEvent('b', Point2D(x=1)),
    ])
    traces.append([
        TriggerEvent(Point2D(x=1), MonitorState.ACTIVE),
        SpamEvent('b', Point2D()),
        SpamEvent('b', Point2D()),
        SpamEvent('b', Point2D(x=1), state=MonitorState.SAFE),
        SpamEvent('b', Point2D(x=1)),
    ])
    # invalid
    traces.append([
        TriggerEvent(Point2D(x=1), MonitorState.ACTIVE),
        TimerEvent(),
        BehaviourEvent(Point2D(x=1), MonitorState.FALSE),
        TimerEvent(),
        SpamEvent('a', Point2D(x=2)),
        SpamEvent('b', Point2D(x=2)),
    ])
    traces.append([
        TriggerEvent(Point2D(x=1), MonitorState.ACTIVE),
        TimerEvent(),
        TimerEvent(),
        TimerEvent(state=MonitorState.SAFE),
        TriggerEvent(Point2D(x=1), MonitorState.ACTIVE),
        TriggerEvent(Point2D(x=2), None),
        BehaviourEvent(Point2D(x=1), MonitorState.FALSE),
    ])
    return (text, traces)

def globally_forbids_ref_within():
    text = 'globally: a as A {x > 0} forbids b {x > 0 and x > @A.x} within 3 s'
    traces = []
    # valid
    traces.append([])
    traces.append([ TimerEvent() ])
    traces.append([ SpamEvent('a', Point2D()) ])
    traces.append([ SpamEvent('b', Point2D(x=1)) ])
    traces.append([
        SpamEvent('b', Point2D(x=1)),
        TimerEvent(),
        SpamEvent('b', Point2D(x=2)),
    ])
    traces.append([
        TriggerEvent(Point2D(x=1), MonitorState.ACTIVE),
        TimerEvent(),
        TimerEvent(),
        TimerEvent(state=MonitorState.SAFE),
        SpamEvent('b', Point2D(x=2)),
    ])
    traces.append([
        TriggerEvent(Point2D(x=1), MonitorState.ACTIVE),
        SpamEvent('b', Point2D(x=1)),
        SpamEvent('b', Point2D(x=1)),
        SpamEvent('b', Point2D(x=1), state=MonitorState.SAFE),
        SpamEvent('b', Point2D(x=2)),
    ])
    # invalid
    traces.append([
        TriggerEvent(Point2D(x=1), MonitorState.ACTIVE),
        TimerEvent(),
        BehaviourEvent(Point2D(x=2), MonitorState.FALSE),
        TimerEvent(),
        SpamEvent('a', Point2D(x=2)),
        SpamEvent('b', Point2D(x=3)),
    ])
    traces.append([
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        SpamEvent('b', Point2D(x=2)),
        SpamEvent('b', Point2D(x=2)),
        SpamEvent('b', Point2D(x=2), state=MonitorState.SAFE),
        TriggerEvent(Point2D(x=1), MonitorState.ACTIVE),
        BehaviourEvent(Point2D(x=2), MonitorState.FALSE),
        SpamEvent('b', Point2D(x=3)),
    ])
    return (text, traces)

###############################################################################
# After Scope No Timeout
###############################################################################

def after_forbids():
    text = 'after p as P {x > 0}: a {x > @P.x} forbids b {x > @P.x}'
    traces = []
    # valid
    traces.append([])
    traces.append([ TimerEvent() ])
    traces.append([ SpamEvent('b', Point2D(x=1)) ])
    traces.append([ SpamEvent('a', Point2D(x=1)) ])
    traces.append([ SpamEvent('p', Point2D()) ])
    traces.append([ ActivatorEvent(Point2D(x=1), MonitorState.SAFE) ])
    # invalid
    traces.append([
        SpamEvent('b', Point2D(x=1)),
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        TimerEvent(),
        SpamEvent('p', Point2D(x=1)),
        SpamEvent('b', Point2D(x=2)),
        SpamEvent('a', Point2D(x=1)),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        SpamEvent('a', Point2D(x=3)),
        BehaviourEvent(Point2D(x=2), MonitorState.FALSE),
    ])
    return (text, traces)

def after_forbids_ref():
    text = 'after p as P {x > 0}: a as A {x > @P.x} forbids b {x > @P.x and x > @A.x}'
    traces = []
    # valid
    traces.append([])
    traces.append([ TimerEvent() ])
    traces.append([ SpamEvent('b', Point2D(x=1)) ])
    traces.append([ SpamEvent('a', Point2D(x=1)) ])
    traces.append([ SpamEvent('p', Point2D()) ])
    traces.append([ ActivatorEvent(Point2D(x=1), MonitorState.SAFE) ])
    # invalid
    traces.append([
        SpamEvent('b', Point2D(x=1)),
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        TimerEvent(),
        SpamEvent('p', Point2D(x=2)),
        SpamEvent('b', Point2D(x=2)),
        SpamEvent('a', Point2D(x=1)),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        TriggerEvent(Point2D(x=3), MonitorState.ACTIVE),
        SpamEvent('b', Point2D(x=2)),
        BehaviourEvent(Point2D(x=3), MonitorState.FALSE),
        SpamEvent('a', Point2D(x=3)),
        SpamEvent('b', Point2D(x=4)),
    ])
    return (text, traces)

###############################################################################
# After Scope With Timeout
###############################################################################

def after_forbids_within():
    text = 'after p as P {x > 0}: a {x > @P.x} forbids b {x > @P.x} within 3 s'
    traces = []
    # valid
    traces.append([])
    traces.append([ TimerEvent() ])
    traces.append([ SpamEvent('p', Point2D()) ])
    traces.append([ SpamEvent('a', Point2D(x=1)) ])
    traces.append([ SpamEvent('b', Point2D(x=1)) ])
    traces.append([ ActivatorEvent(Point2D(x=1), MonitorState.SAFE) ])
    traces.append([
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        TimerEvent(),
        TimerEvent(),
        TimerEvent(state=MonitorState.SAFE),
    ])
    traces.append([
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        SpamEvent('b', Point2D(x=1)),
        SpamEvent('b', Point2D(x=1)),
        SpamEvent('b', Point2D(x=2), MonitorState.SAFE),
        SpamEvent('b', Point2D(x=2)),
    ])
    traces.append([
        SpamEvent('b', Point2D()),
        TimerEvent(),
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        TimerEvent(),
        SpamEvent('b', Point2D(x=1)),
        SpamEvent('b', Point2D(x=1), MonitorState.SAFE),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        TimerEvent(),
        TriggerEvent(Point2D(x=2), None),
        SpamEvent('b', Point2D(x=1)),
        SpamEvent('b', Point2D(x=1)),
        SpamEvent('b', Point2D(x=2), state=MonitorState.SAFE),
    ])
    # invalid
    traces.append([
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        BehaviourEvent(Point2D(x=2), MonitorState.FALSE),
    ])
    traces.append([
        SpamEvent('b', Point2D(x=1)),
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        TimerEvent(),
        TimerEvent(),
        SpamEvent('b', Point2D(x=2), state=MonitorState.SAFE),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        TimerEvent(),
        TimerEvent(),
        TriggerEvent(Point2D(x=3), None),
        BehaviourEvent(Point2D(x=2), MonitorState.FALSE),
    ])
    return (text, traces)

def after_forbids_ref_within():
    text = 'after p as P {x > 0}: a as A {x > @P.x} forbids b {x > @P.x and x > @A.x} within 3 s'
    traces = []
    # valid
    traces.append([])
    traces.append([ TimerEvent() ])
    traces.append([ SpamEvent('p', Point2D()) ])
    traces.append([ SpamEvent('a', Point2D(x=1)) ])
    traces.append([ SpamEvent('b', Point2D(x=1)) ])
    traces.append([ ActivatorEvent(Point2D(x=1), MonitorState.SAFE) ])
    traces.append([
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        TimerEvent(),
        TimerEvent(),
        TimerEvent(state=MonitorState.SAFE),
    ])
    traces.append([
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        SpamEvent('b', Point2D()),
        SpamEvent('b', Point2D(x=1)),
        SpamEvent('b', Point2D(x=2), MonitorState.SAFE),
        SpamEvent('b', Point2D(x=3)),
    ])
    traces.append([
        SpamEvent('b', Point2D()),
        TimerEvent(),
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        TimerEvent(),
        SpamEvent('b', Point2D(x=1)),
        SpamEvent('b', Point2D(x=2), state=MonitorState.SAFE),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        TimerEvent(),
        TriggerEvent(Point2D(x=3), None),
        SpamEvent('b', Point2D(x=3)),
        TimerEvent(),
        SpamEvent('b', Point2D(x=3), state=MonitorState.SAFE),
        SpamEvent('b', Point2D(x=4)),
    ])
    # invalid
    traces.append([
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        SpamEvent('a', Point2D(x=1)),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        SpamEvent('b', Point2D(x=2)),
        BehaviourEvent(Point2D(x=3), MonitorState.FALSE),
    ])
    traces.append([
        SpamEvent('a', Point2D(x=1)),
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        SpamEvent('a', Point2D(x=1)),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        SpamEvent('b', Point2D(x=2)),
        SpamEvent('b', Point2D(x=2)),
        SpamEvent('b', Point2D(x=2), state=MonitorState.SAFE),
        SpamEvent('b', Point2D(x=3)),
        TimerEvent(),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        SpamEvent('b', Point2D(x=2)),
        BehaviourEvent(Point2D(x=4), MonitorState.FALSE),
        SpamEvent('b', Point2D(x=4)),
    ])
    traces.append([
        SpamEvent('a', Point2D(x=1)),
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        SpamEvent('a', Point2D(x=1)),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        SpamEvent('b', Point2D(x=2)),
        TriggerEvent(Point2D(x=3), MonitorState.ACTIVE),
        SpamEvent('b', Point2D(x=3)),
        BehaviourEvent(Point2D(x=4), MonitorState.FALSE),
    ])
    return (text, traces)

###############################################################################
# Until Scope No Timeout
###############################################################################

def until_forbids():
    text = 'until q {x > 0}: a {x > 0} forbids b {x > 0}'
    traces = []
    # valid
    traces.append([])
    traces.append([ TimerEvent() ])
    traces.append([ SpamEvent('b', Point2D(x=1)) ])
    traces.append([ TerminatorEvent(Point2D(x=1), MonitorState.TRUE) ])
    traces.append([ TriggerEvent(Point2D(x=1), MonitorState.ACTIVE) ])
    traces.append([
        SpamEvent('b', Point2D(x=1)),
        TimerEvent(),
        SpamEvent('b', Point2D(x=2)),
    ])
    traces.append([
        TriggerEvent(Point2D(x=1), MonitorState.ACTIVE),
        TerminatorEvent(Point2D(x=1), MonitorState.TRUE),
    ])
    traces.append([
        SpamEvent('b', Point2D(x=1)),
        TimerEvent(),
        TriggerEvent(Point2D(x=1), MonitorState.ACTIVE),
        SpamEvent('b', Point2D(x=0)),
        SpamEvent('a', Point2D(x=2)),
        TerminatorEvent(Point2D(x=1), MonitorState.TRUE),
        SpamEvent('a', Point2D(x=2)),
        SpamEvent('b', Point2D(x=2)),
        SpamEvent('q', Point2D(x=2)),
    ])
    # invalid
    traces.append([
        TriggerEvent(Point2D(x=1), MonitorState.ACTIVE),
        BehaviourEvent(Point2D(x=1), MonitorState.FALSE),
    ])
    traces.append([
        SpamEvent('b', Point2D(x=1)),
        TimerEvent(),
        TriggerEvent(Point2D(x=1), MonitorState.ACTIVE),
        SpamEvent('a', Point2D(x=2)),
        SpamEvent('b', Point2D(x=0)),
        BehaviourEvent(Point2D(x=1), MonitorState.FALSE),
        SpamEvent('a', Point2D(x=2)),
        SpamEvent('q', Point2D(x=2)),
        SpamEvent('b', Point2D(x=2)),
    ])
    return (text, traces)

def until_forbids_ref():
    text = 'until q {x > 0}: a as A {x > 0} forbids b {x > 0 and x > @A.x}'
    traces = []
    # valid
    traces.append([])
    traces.append([ TimerEvent() ])
    traces.append([ SpamEvent('b', Point2D(x=1)) ])
    traces.append([ TerminatorEvent(Point2D(x=1), MonitorState.TRUE) ])
    traces.append([ TriggerEvent(Point2D(x=1), MonitorState.ACTIVE) ])
    traces.append([
        SpamEvent('b', Point2D(x=1)),
        TimerEvent(),
        SpamEvent('b', Point2D(x=2)),
    ])
    traces.append([
        TriggerEvent(Point2D(x=1), MonitorState.ACTIVE),
        SpamEvent('b', Point2D(x=1)),
        TerminatorEvent(Point2D(x=1), MonitorState.TRUE),
    ])
    traces.append([
        SpamEvent('b', Point2D(x=1)),
        TimerEvent(),
        TriggerEvent(Point2D(x=1), MonitorState.ACTIVE),
        TriggerEvent(Point2D(x=2), None),
        SpamEvent('b', Point2D(x=1)),
        TerminatorEvent(Point2D(x=1), MonitorState.TRUE),
        SpamEvent('a', Point2D(x=2)),
        SpamEvent('b', Point2D(x=3)),
        SpamEvent('q', Point2D(x=2)),
    ])
    # invalid
    traces.append([
        TriggerEvent(Point2D(x=1), MonitorState.ACTIVE),
        BehaviourEvent(Point2D(x=2), MonitorState.FALSE),
    ])
    traces.append([
        SpamEvent('b', Point2D(x=1)),
        TimerEvent(),
        TriggerEvent(Point2D(x=3), MonitorState.ACTIVE),
        SpamEvent('b', Point2D(x=2)),
        TriggerEvent(Point2D(x=1), None),
        BehaviourEvent(Point2D(x=2), MonitorState.FALSE),
        SpamEvent('a', Point2D(x=2)),
        SpamEvent('b', Point2D(x=2)),
        SpamEvent('q', Point2D(x=2)),
    ])
    return (text, traces)

###############################################################################
# Until Scope With Timeout
###############################################################################

def until_forbids_within():
    text = 'until q {x > 0}: a {x > 0} forbids b {x > 0} within 3 s'
    traces = []
    # valid
    traces.append([])
    traces.append([ TimerEvent() ])
    traces.append([ SpamEvent('b', Point2D(x=1)) ])
    traces.append([ TerminatorEvent(Point2D(x=1), MonitorState.TRUE) ])
    traces.append([ TriggerEvent(Point2D(x=1), MonitorState.ACTIVE) ])
    traces.append([
        SpamEvent('b', Point2D(x=1)),
        SpamEvent('a', Point2D()),
        TimerEvent(),
        SpamEvent('b', Point2D(x=2)),
        SpamEvent('q', Point2D()),
    ])
    traces.append([
        SpamEvent('b', Point2D()),
        TimerEvent(),
        TerminatorEvent(Point2D(x=1), MonitorState.TRUE),
        SpamEvent('a', Point2D(x=2)),
        SpamEvent('b', Point2D(x=2)),
        SpamEvent('q', Point2D(x=2)),
    ])
    traces.append([
        TriggerEvent(Point2D(x=1), MonitorState.ACTIVE),
        TerminatorEvent(Point2D(x=1), MonitorState.TRUE),
    ])
    traces.append([
        TriggerEvent(Point2D(x=1), MonitorState.ACTIVE),
        SpamEvent('a', Point2D()),
        SpamEvent('b', Point2D()),
        SpamEvent('b', Point2D(x=2), state=MonitorState.SAFE),
        TerminatorEvent(Point2D(x=1), MonitorState.TRUE),
    ])
    traces.append([
        TriggerEvent(Point2D(x=1), MonitorState.ACTIVE),
        TimerEvent(),
        TimerEvent(),
        TimerEvent(state=MonitorState.SAFE),
    ])
    traces.append([
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        SpamEvent('a', Point2D(x=0)),
        SpamEvent('b', Point2D(x=0)),
        SpamEvent('b', Point2D(x=0), state=MonitorState.SAFE),
        TriggerEvent(Point2D(x=1), MonitorState.ACTIVE),
        SpamEvent('a', Point2D(x=0)),
        SpamEvent('b', Point2D(x=0)),
        SpamEvent('b', Point2D(x=0), state=MonitorState.SAFE),
    ])
    # invalid
    traces.append([
        SpamEvent('b', Point2D()),
        TimerEvent(),
        TriggerEvent(Point2D(x=1), MonitorState.ACTIVE),
        TriggerEvent(Point2D(x=2), None),
        BehaviourEvent(Point2D(x=2), MonitorState.FALSE),
        SpamEvent('a', Point2D(x=2)),
        SpamEvent('b', Point2D(x=2)),
        SpamEvent('q', Point2D(x=2)),
    ])
    traces.append([
        TriggerEvent(Point2D(x=1), MonitorState.ACTIVE),
        SpamEvent('b', Point2D(x=0)),
        SpamEvent('b', Point2D(x=0)),
        SpamEvent('b', Point2D(x=0), state=MonitorState.SAFE),
        TriggerEvent(Point2D(x=1), MonitorState.ACTIVE),
        BehaviourEvent(Point2D(x=2), MonitorState.FALSE),
    ])
    traces.append([
        TriggerEvent(Point2D(x=1), MonitorState.ACTIVE),
        SpamEvent('b', Point2D(x=0)),
        TriggerEvent(Point2D(x=1), None),
        SpamEvent('b', Point2D(x=0)),
        BehaviourEvent(Point2D(x=2), MonitorState.FALSE),
    ])
    return (text, traces)

def until_forbids_ref_within():
    text = 'until q {x > 0}: a as A {x > 0} forbids b {x > 0 and x > @A.x} within 3 s'
    traces = []
    # valid
    traces.append([])
    traces.append([ TimerEvent() ])
    traces.append([ SpamEvent('b', Point2D(x=1)) ])
    traces.append([ TerminatorEvent(Point2D(x=1), MonitorState.TRUE) ])
    traces.append([ TriggerEvent(Point2D(x=1), MonitorState.ACTIVE) ])
    traces.append([
        SpamEvent('b', Point2D(x=1)),
        SpamEvent('a', Point2D()),
        TimerEvent(),
        SpamEvent('b', Point2D(x=2)),
        SpamEvent('q', Point2D()),
    ])
    traces.append([
        SpamEvent('b', Point2D()),
        TimerEvent(),
        TerminatorEvent(Point2D(x=1), MonitorState.TRUE),
        SpamEvent('a', Point2D(x=2)),
        SpamEvent('b', Point2D(x=2)),
        SpamEvent('q', Point2D(x=2)),
    ])
    traces.append([
        TriggerEvent(Point2D(x=1), MonitorState.ACTIVE),
        TerminatorEvent(Point2D(x=2), MonitorState.TRUE),
    ])
    traces.append([
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        TriggerEvent(Point2D(x=1), None),
        SpamEvent('b', Point2D(x=1), None),
        SpamEvent('b', Point2D(x=1), None),
        SpamEvent('b', Point2D(x=3), state=MonitorState.SAFE),
        TerminatorEvent(Point2D(x=1), MonitorState.TRUE),
    ])
    traces.append([
        TriggerEvent(Point2D(x=1), MonitorState.ACTIVE),
        TimerEvent(),
        TimerEvent(),
        TimerEvent(state=MonitorState.SAFE),
    ])
    traces.append([
        TriggerEvent(Point2D(x=3), MonitorState.ACTIVE),
        SpamEvent('b', Point2D(x=1)),
        SpamEvent('b', Point2D(x=2)),
        SpamEvent('b', Point2D(x=3), state=MonitorState.SAFE),
        TriggerEvent(Point2D(x=1), MonitorState.ACTIVE),
        TerminatorEvent(Point2D(x=2), MonitorState.TRUE),
    ])
    # invalid
    traces.append([
        TriggerEvent(Point2D(x=1), MonitorState.ACTIVE),
        BehaviourEvent(Point2D(x=2), MonitorState.FALSE),
    ])
    traces.append([
        TriggerEvent(Point2D(x=3), MonitorState.ACTIVE),
        SpamEvent('b', Point2D(x=1)),
        SpamEvent('b', Point2D(x=2)),
        SpamEvent('b', Point2D(x=3), state=MonitorState.SAFE),
        TriggerEvent(Point2D(x=1), MonitorState.ACTIVE),
        BehaviourEvent(Point2D(x=2), MonitorState.FALSE),
    ])
    traces.append([
        SpamEvent('b', Point2D(x=2)),
        TimerEvent(),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        SpamEvent('b', Point2D(x=2)),
        TriggerEvent(Point2D(x=1), None),
        BehaviourEvent(Point2D(x=2), MonitorState.FALSE),
        TimerEvent(),
        SpamEvent('a', Point2D(x=2)),
        SpamEvent('b', Point2D(x=2)),
        SpamEvent('q', Point2D(x=2)),
    ])
    return (text, traces)

###############################################################################
# After-Until Scope No Timeout
###############################################################################

def after_until_forbids():
    text = 'after p as P {x > 0} until q {x > @P.x}: a {x > @P.x} forbids b {x > @P.x}'
    traces = []
    # valid
    traces.append([])
    traces.append([ TimerEvent() ])
    traces.append([ SpamEvent('b', Point2D(x=1)) ])
    traces.append([ ActivatorEvent(Point2D(x=1), MonitorState.SAFE) ])
    traces.append([
        SpamEvent('a', Point2D(x=1)),
        SpamEvent('b', Point2D(x=1)),
        SpamEvent('q', Point2D(x=1)),
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        TimerEvent(),
        SpamEvent('b', Point2D(x=2)),
    ])
    traces.append([
        SpamEvent('p', Point2D()),
        SpamEvent('q', Point2D(x=1)),
        SpamEvent('b', Point2D(x=1)),
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        SpamEvent('a', Point2D(x=1)),
        SpamEvent('q', Point2D(x=1)),
        TerminatorEvent(Point2D(x=2), MonitorState.INACTIVE),
        SpamEvent('a', Point2D(x=2)),
        SpamEvent('q', Point2D(x=2)),
        ActivatorEvent(Point2D(x=2), MonitorState.SAFE),
        SpamEvent('a', Point2D(x=2)),
        SpamEvent('q', Point2D(x=2)),
        TerminatorEvent(Point2D(x=3), MonitorState.INACTIVE),
    ])
    traces.append([
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        SpamEvent('b', Point2D(x=2)),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        SpamEvent('a', Point2D(x=3)),
        SpamEvent('b', Point2D(x=1)),
        SpamEvent('q', Point2D(x=1)),
        TerminatorEvent(Point2D(x=2), MonitorState.INACTIVE),
        SpamEvent('a', Point2D(x=3)),
        SpamEvent('b', Point2D(x=3)),
        SpamEvent('q', Point2D(x=3)),
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        SpamEvent('b', Point2D(x=2)),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        SpamEvent('a', Point2D(x=3)),
        SpamEvent('b', Point2D(x=1)),
        SpamEvent('q', Point2D(x=1)),
        TerminatorEvent(Point2D(x=2), MonitorState.INACTIVE),
    ])
    # invalid
    traces.append([
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        BehaviourEvent(Point2D(x=2), MonitorState.FALSE),
        SpamEvent('a', Point2D(x=3)),
        SpamEvent('b', Point2D(x=3)),
        SpamEvent('q', Point2D(x=3)),
        SpamEvent('p', Point2D(x=3)),
    ])
    traces.append([
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        SpamEvent('b', Point2D(x=2)),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        SpamEvent('a', Point2D(x=3)),
        SpamEvent('b', Point2D(x=1)),
        SpamEvent('q', Point2D(x=1)),
        TerminatorEvent(Point2D(x=2), MonitorState.INACTIVE),
        SpamEvent('a', Point2D(x=3)),
        SpamEvent('b', Point2D(x=3)),
        SpamEvent('q', Point2D(x=3)),
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        SpamEvent('b', Point2D(x=2)),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        SpamEvent('a', Point2D(x=3)),
        BehaviourEvent(Point2D(x=3), MonitorState.FALSE),
        SpamEvent('q', Point2D(x=2)),
    ])
    return (text, traces)

def after_until_forbids_ref():
    text = 'after p as P {x > 0} until q {x > @P.x}: a as A {x > @P.x} forbids b {x > @P.x and x > @A.x}'
    traces = []
    # valid
    traces.append([])
    traces.append([ TimerEvent() ])
    traces.append([ SpamEvent('b', Point2D(x=1)) ])
    traces.append([ ActivatorEvent(Point2D(x=1), MonitorState.SAFE) ])
    traces.append([
        SpamEvent('a', Point2D(x=1)),
        SpamEvent('b', Point2D(x=1)),
        SpamEvent('q', Point2D(x=1)),
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        TimerEvent(),
        SpamEvent('b', Point2D(x=2)),
    ])
    traces.append([
        SpamEvent('p', Point2D()),
        SpamEvent('q', Point2D(x=1)),
        SpamEvent('b', Point2D(x=1)),
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        SpamEvent('a', Point2D(x=1)),
        SpamEvent('q', Point2D(x=1)),
        TerminatorEvent(Point2D(x=2), MonitorState.INACTIVE),
        SpamEvent('a', Point2D(x=2)),
        SpamEvent('q', Point2D(x=2)),
        ActivatorEvent(Point2D(x=2), MonitorState.SAFE),
        SpamEvent('a', Point2D(x=2)),
        SpamEvent('q', Point2D(x=2)),
        TerminatorEvent(Point2D(x=3), MonitorState.INACTIVE),
    ])
    traces.append([
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        SpamEvent('b', Point2D(x=2)),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        TriggerEvent(Point2D(x=3), None),
        SpamEvent('b', Point2D(x=2)),
        SpamEvent('q', Point2D(x=1)),
        TerminatorEvent(Point2D(x=2), MonitorState.INACTIVE),
        SpamEvent('a', Point2D(x=3)),
        SpamEvent('b', Point2D(x=3)),
        SpamEvent('q', Point2D(x=3)),
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        SpamEvent('b', Point2D(x=2)),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        TriggerEvent(Point2D(x=3), None),
        SpamEvent('b', Point2D(x=2)),
        SpamEvent('q', Point2D(x=1)),
        TerminatorEvent(Point2D(x=2), MonitorState.INACTIVE),
    ])
    # invalid
    traces.append([
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        BehaviourEvent(Point2D(x=3), MonitorState.FALSE),
        SpamEvent('a', Point2D(x=3)),
        SpamEvent('b', Point2D(x=3)),
        SpamEvent('q', Point2D(x=3)),
        SpamEvent('p', Point2D(x=3)),
    ])
    traces.append([
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        SpamEvent('b', Point2D(x=2)),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        TriggerEvent(Point2D(x=3), None),
        SpamEvent('b', Point2D(x=2)),
        SpamEvent('q', Point2D(x=1)),
        TerminatorEvent(Point2D(x=2), MonitorState.INACTIVE),
        SpamEvent('a', Point2D(x=3)),
        SpamEvent('b', Point2D(x=3)),
        SpamEvent('q', Point2D(x=3)),
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        SpamEvent('b', Point2D(x=2)),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        TriggerEvent(Point2D(x=3), None),
        BehaviourEvent(Point2D(x=3), MonitorState.FALSE),
        SpamEvent('q', Point2D(x=2)),
    ])
    traces.append([
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        SpamEvent('b', Point2D(x=3)),
        TriggerEvent(Point2D(x=3), MonitorState.ACTIVE),
        SpamEvent('b', Point2D(x=3)),
        TriggerEvent(Point2D(x=2), None),
        BehaviourEvent(Point2D(x=3), MonitorState.FALSE),
    ])
    return (text, traces)

###############################################################################
# After-Until Scope With Timeout
###############################################################################

def after_until_forbids_within():
    text = 'after p as P {x > 0} until q {x > @P.x}: a {x > @P.x} forbids b {x > @P.x} within 3 s'
    traces = []
    # valid
    traces.append([])
    traces.append([ TimerEvent() ])
    traces.append([ SpamEvent('b', Point2D(x=1)) ])
    traces.append([ ActivatorEvent(Point2D(x=1), MonitorState.SAFE) ])
    traces.append([
        SpamEvent('a', Point2D(x=1)),
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        TimerEvent(),
        SpamEvent('a', Point2D(x=1)),
    ])
    traces.append([
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        TerminatorEvent(Point2D(x=2), MonitorState.INACTIVE),
    ])
    traces.append([
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        TimerEvent(),
        TimerEvent(),
        TimerEvent(state=MonitorState.SAFE),
    ])
    traces.append([
        SpamEvent('b', Point2D(x=2)),
        TimerEvent(),
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        SpamEvent('b', Point2D(x=1)),
        TerminatorEvent(Point2D(x=2), MonitorState.INACTIVE),
        TimerEvent(),
        SpamEvent('b', Point2D(x=3)),
    ])
    traces.append([
        SpamEvent('p', Point2D()),
        SpamEvent('q', Point2D(x=2)),
        SpamEvent('a', Point2D(x=2)),
        SpamEvent('b', Point2D(x=2)),
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        SpamEvent('a', Point2D(x=1)),
        SpamEvent('b', Point2D(x=1)),
        SpamEvent('q', Point2D(x=1)),
        TerminatorEvent(Point2D(x=2), MonitorState.INACTIVE),
        SpamEvent('a', Point2D(x=2)),
        SpamEvent('b', Point2D(x=2)),
        SpamEvent('q', Point2D(x=2)),
        ActivatorEvent(Point2D(x=2), MonitorState.SAFE),
        TriggerEvent(Point2D(x=3), MonitorState.ACTIVE),
        TerminatorEvent(Point2D(x=3), MonitorState.INACTIVE),
        SpamEvent('b', Point2D(x=3)),
        SpamEvent('q', Point2D(x=3)),
    ])
    traces.append([
        SpamEvent('p', Point2D()),
        SpamEvent('q', Point2D(x=1)),
        SpamEvent('a', Point2D(x=1)),
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        SpamEvent('b', Point2D(x=2)),
        SpamEvent('q', Point2D(x=1)),
        TerminatorEvent(Point2D(x=2), MonitorState.INACTIVE),
        SpamEvent('a', Point2D(x=2)),
        SpamEvent('q', Point2D(x=2)),
        ActivatorEvent(Point2D(x=2), MonitorState.SAFE),
        SpamEvent('a', Point2D(x=2)),
        SpamEvent('q', Point2D(x=2)),
        TerminatorEvent(Point2D(x=3), MonitorState.INACTIVE),
    ])
    traces.append([
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        SpamEvent('b', Point2D(x=1)),
        TerminatorEvent(Point2D(x=2), MonitorState.INACTIVE),
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        SpamEvent('b', Point2D(x=1)),
        TimerEvent(),
        TerminatorEvent(Point2D(x=2), MonitorState.INACTIVE),
        SpamEvent('a', Point2D(x=2)),
        SpamEvent('b', Point2D(x=3)),
        SpamEvent('q', Point2D(x=2)),
        ActivatorEvent(Point2D(x=2), MonitorState.SAFE),
        TriggerEvent(Point2D(x=4), MonitorState.ACTIVE),
        SpamEvent('b', Point2D(x=2)),
        SpamEvent('b', Point2D(x=2)),
        SpamEvent('b', Point2D(x=2), state=MonitorState.SAFE),
        TriggerEvent(Point2D(x=3), MonitorState.ACTIVE),
        SpamEvent('b', Point2D(x=2)),
        TerminatorEvent(Point2D(x=3), MonitorState.INACTIVE),
    ])
    # invalid
    traces.append([
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        TriggerEvent(Point2D(x=3), None),
        BehaviourEvent(Point2D(x=3), MonitorState.FALSE),
        SpamEvent('a', Point2D(x=3)),
        SpamEvent('b', Point2D(x=3)),
        SpamEvent('p', Point2D(x=3)),
        SpamEvent('q', Point2D(x=3)),
    ])
    traces.append([
        SpamEvent('p', Point2D()),
        SpamEvent('q', Point2D(x=1)),
        SpamEvent('a', Point2D(x=1)),
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        SpamEvent('b', Point2D(x=2)),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        TriggerEvent(Point2D(x=3), None),
        SpamEvent('b', Point2D(x=1)),
        SpamEvent('b', Point2D(x=1)),
        SpamEvent('b', Point2D(x=1), state=MonitorState.SAFE),
        SpamEvent('b', Point2D(x=3)),
        SpamEvent('q', Point2D(x=1)),
        TerminatorEvent(Point2D(x=2), MonitorState.INACTIVE),
        SpamEvent('a', Point2D(x=3)),
        SpamEvent('b', Point2D(x=3)),
        SpamEvent('q', Point2D(x=3)),
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        SpamEvent('b', Point2D(x=2)),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        TriggerEvent(Point2D(x=3), None),
        SpamEvent('b', Point2D(x=1)),
        BehaviourEvent(Point2D(x=2), MonitorState.FALSE),
        SpamEvent('b', Point2D(x=2)),
        TimerEvent(),
        SpamEvent('q', Point2D(x=2)),
    ])
    return (text, traces)

def after_until_forbids_ref_within():
    text = 'after p as P {x > 0} until q {x > @P.x}: a as A {x > @P.x} forbids b {x > @P.x and x > @A.x} within 3 s'
    traces = []
    # valid
    traces.append([])
    traces.append([ TimerEvent() ])
    traces.append([ SpamEvent('b', Point2D(x=1)) ])
    traces.append([ ActivatorEvent(Point2D(x=1), MonitorState.SAFE) ])
    traces.append([
        SpamEvent('a', Point2D(x=1)),
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        TimerEvent(),
        SpamEvent('a', Point2D(x=1)),
    ])
    traces.append([
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        SpamEvent('b', Point2D(x=2)),
        TriggerEvent(Point2D(x=3), None),
        SpamEvent('b', Point2D(x=2)),
        SpamEvent('b', Point2D(x=3)),
        SpamEvent('b', Point2D(x=4), state=MonitorState.SAFE),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        TriggerEvent(Point2D(x=3), None),
        SpamEvent('b', Point2D(x=2)),
        SpamEvent('b', Point2D(x=3)),
        SpamEvent('b', Point2D(x=4), state=MonitorState.SAFE),
    ])
    traces.append([
        SpamEvent('p', Point2D()),
        SpamEvent('q', Point2D(x=1)),
        SpamEvent('a', Point2D(x=1)),
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        SpamEvent('b', Point2D(x=2)),
        SpamEvent('q', Point2D(x=1)),
        TerminatorEvent(Point2D(x=2), MonitorState.INACTIVE),
        SpamEvent('a', Point2D(x=2)),
        SpamEvent('q', Point2D(x=2)),
        ActivatorEvent(Point2D(x=2), MonitorState.SAFE),
        SpamEvent('a', Point2D(x=2)),
        SpamEvent('q', Point2D(x=2)),
        TerminatorEvent(Point2D(x=3), MonitorState.INACTIVE),
    ])
    traces.append([
        SpamEvent('p', Point2D()),
        SpamEvent('q', Point2D(x=1)),
        SpamEvent('a', Point2D(x=1)),
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        SpamEvent('b', Point2D(x=2)),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        TriggerEvent(Point2D(x=3), None),
        TerminatorEvent(Point2D(x=2), MonitorState.INACTIVE),
        SpamEvent('a', Point2D(x=3)),
        SpamEvent('b', Point2D(x=3)),
        SpamEvent('q', Point2D(x=3)),
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        SpamEvent('b', Point2D(x=2)),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        TriggerEvent(Point2D(x=3), None),
        SpamEvent('b', Point2D(x=2)),
        TimerEvent(),
        TerminatorEvent(Point2D(x=2), MonitorState.INACTIVE),
    ])
    # invalid
    traces.append([
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        BehaviourEvent(Point2D(x=3), MonitorState.FALSE),
    ])
    traces.append([
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        TimerEvent(),
        TimerEvent(),
        TimerEvent(state=MonitorState.SAFE),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        BehaviourEvent(Point2D(x=3), MonitorState.FALSE),
    ])
    traces.append([
        SpamEvent('b', Point2D(x=2)),
        TimerEvent(),
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        SpamEvent('b', Point2D(x=2)),
        TerminatorEvent(Point2D(x=2), MonitorState.INACTIVE),
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        TimerEvent(),
        BehaviourEvent(Point2D(x=3), MonitorState.FALSE),
    ])
    traces.append([
        SpamEvent('p', Point2D()),
        SpamEvent('q', Point2D(x=2)),
        SpamEvent('a', Point2D(x=2)),
        SpamEvent('b', Point2D(x=2)),
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        SpamEvent('a', Point2D(x=1)),
        SpamEvent('b', Point2D(x=1)),
        SpamEvent('q', Point2D(x=1)),
        TerminatorEvent(Point2D(x=2), MonitorState.INACTIVE),
        SpamEvent('a', Point2D(x=2)),
        SpamEvent('b', Point2D(x=2)),
        SpamEvent('q', Point2D(x=2)),
        ActivatorEvent(Point2D(x=2), MonitorState.SAFE),
        TriggerEvent(Point2D(x=3), MonitorState.ACTIVE),
        BehaviourEvent(Point2D(x=4), MonitorState.FALSE),
        SpamEvent('b', Point2D(x=4)),
        SpamEvent('q', Point2D(x=4)),
    ])
    traces.append([
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        TerminatorEvent(Point2D(x=2), MonitorState.INACTIVE),
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        TriggerEvent(Point2D(x=2), MonitorState.ACTIVE),
        SpamEvent('b', Point2D(x=2)),
        TimerEvent(),
        SpamEvent('b', Point2D(x=2), state=MonitorState.SAFE),
        SpamEvent('b', Point2D(x=3)),
        TerminatorEvent(Point2D(x=2), MonitorState.INACTIVE),
        SpamEvent('a', Point2D(x=2)),
        SpamEvent('b', Point2D(x=3)),
        SpamEvent('q', Point2D(x=2)),
        ActivatorEvent(Point2D(x=1), MonitorState.SAFE),
        TriggerEvent(Point2D(x=4), MonitorState.ACTIVE),
        SpamEvent('b', Point2D(x=3)),
        TriggerEvent(Point2D(x=2), None),
        BehaviourEvent(Point2D(x=3), MonitorState.FALSE),
    ])
    return (text, traces)



# SPDX-License-Identifier: MIT
# Copyright © 2023 André Santos

###############################################################################
# Imports
###############################################################################

from typing import Optional

from collections import defaultdict
from enum import IntEnum

from attrs import frozen
from hpl.ast import HplPredicate, HplVacuousTruth
from hpl.rewrite import refactor_reference, replace_this_with_var, replace_var_with_this

###############################################################################
# Constants
###############################################################################

INF = float('inf')


class MonitorState(IntEnum):
    OFF = 0
    TRUE = -1
    FALSE = -2
    INACTIVE = 1
    ACTIVE = 2
    SAFE = 3


class EventType(IntEnum):
    TIMER = 0
    ACTIVATOR = 1
    TERMINATOR = 2
    BEHAVIOUR = 3
    TRIGGER = 4
    SPAM = 5


###############################################################################
# Data Structures
###############################################################################


@frozen
class MonitoringEvent:
    predicate: HplPredicate

    @property
    def event_type(self) -> int:
        raise NotImplementedError()


@frozen
class ActivatorEvent(MonitoringEvent):
    @property
    def event_type(self) -> int:
        return EventType.ACTIVATOR


@frozen
class TerminatorEvent(MonitoringEvent):
    activator: Optional[str] = None
    verdict: Optional[bool] = None

    @property
    def event_type(self) -> int:
        return EventType.TERMINATOR


@frozen
class BehaviourEvent(MonitoringEvent):
    activator: Optional[str] = None
    trigger: Optional[str] = None

    @property
    def event_type(self) -> int:
        return EventType.BEHAVIOUR


@frozen
class TriggerEvent(MonitoringEvent):
    activator: Optional[str] = None

    @property
    def event_type(self) -> int:
        return EventType.TRIGGER


def _default_dict_of_lists():
    return defaultdict(list)


###############################################################################
# State Machine Builder
###############################################################################


class PatternBasedBuilder:
    # initial_state: int
    # timeout: float
    # reentrant_scope: bool
    # pool_size: -1|0|int
    # on_msg:
    #    <topic>:
    #        <state>:
    #            - <event>

    def __init__(self, hpl_property, s0):
        self.property_id = hpl_property.metadata.get('id')
        self.property_title = hpl_property.metadata.get('title')
        self.property_desc = hpl_property.metadata.get('description')
        self.property_text = str(hpl_property)
        self.class_name = 'PropertyMonitor'
        self._activator = None
        self._trigger = None
        self.reentrant_scope = False
        self.timeout = hpl_property.pattern.max_time
        if self.timeout == INF:
            self.timeout = -1
        event = hpl_property.scope.activator
        self.launch_enters_scope = event is None
        if event is not None and event.is_simple_event:
            self._activator = event.alias
        event = hpl_property.pattern.trigger
        if event is not None and event.is_simple_event:
            self._trigger = event.alias
        self.pool_size = self.calc_pool_size(hpl_property)
        self.on_msg = defaultdict(_default_dict_of_lists)
        if hpl_property.scope.is_global:
            self.initial_state = s0
        elif hpl_property.scope.is_after and hpl_property.scope.is_until:
            self.initial_state = MonitorState.INACTIVE
            self.reentrant_scope = True
            self.add_activator(hpl_property.scope.activator)
            self.add_terminator(hpl_property.scope.terminator)
        elif hpl_property.scope.is_after:
            self.initial_state = MonitorState.INACTIVE
            self.add_activator(hpl_property.scope.activator)
        elif hpl_property.scope.is_until:
            self.initial_state = s0
            self.add_terminator(hpl_property.scope.terminator)
        else:
            raise ValueError('unknown scope: ' + str(hpl_property.scope))
        self.add_behaviour(hpl_property.pattern.behaviour)
        if hpl_property.pattern.trigger is not None:
            self.add_trigger(hpl_property.pattern.trigger)

    def add_activator(self, event):
        # must be called before all others
        # assuming only disjunctions or simple events
        for e in event.simple_events():
            datum = ActivatorEvent(e.predicate)
            self.on_msg[e.name][MonitorState.INACTIVE].append(datum)

    def add_terminator(self, event):
        raise NotImplementedError()

    def add_trigger(self, event):
        raise NotImplementedError()

    def add_behaviour(self, event):
        raise NotImplementedError()

    def calc_pool_size(self, hpl_property):
        raise NotImplementedError()


###############################################################################
# Absence State Machine
###############################################################################


class AbsenceBuilder(PatternBasedBuilder):
    def __init__(self, hpl_property):
        super().__init__(hpl_property, MonitorState.ACTIVE)

    @property
    def has_safe_state(self):
        return self.timeout >= 0 and self.timeout < INF and self.reentrant_scope

    def calc_pool_size(self, hpl_property):
        return 0

    def add_terminator(self, event):
        # must be called before pattern events
        v = None if self.reentrant_scope else True
        for e in event.simple_events():
            alias = None
            if self._activator and e.contains_reference(self._activator):
                alias = self._activator
            datum = TerminatorEvent(e.predicate, alias, v)
            states = self.on_msg[e.name]
            states[MonitorState.ACTIVE].append(datum)
            if self.has_safe_state:
                states[MonitorState.SAFE].append(datum)

    def add_behaviour(self, event):
        for e in event.simple_events():
            alias = None
            if self._activator and e.contains_reference(self._activator):
                alias = self._activator
            datum = BehaviourEvent(e.predicate, alias, None)
            self.on_msg[e.name][MonitorState.ACTIVE].append(datum)


###############################################################################
# Existence State Machine
###############################################################################


class ExistenceBuilder(PatternBasedBuilder):
    def __init__(self, hpl_property):
        super().__init__(hpl_property, MonitorState.ACTIVE)

    def calc_pool_size(self, hpl_property):
        return 0

    def add_terminator(self, event):
        # must be called before pattern events
        self.has_safe_state = True
        for e in event.simple_events():
            alias = None
            if self._activator and e.contains_reference(self._activator):
                alias = self._activator
            states = self.on_msg[e.name]
            datum = TerminatorEvent(e.predicate, alias, False)
            states[MonitorState.ACTIVE].append(datum)
            if self.reentrant_scope:
                datum = TerminatorEvent(e.predicate, alias, None)
                states[MonitorState.SAFE].append(datum)

    def add_behaviour(self, event):
        for e in event.simple_events():
            alias = None
            if self._activator and e.contains_reference(self._activator):
                alias = self._activator
            datum = BehaviourEvent(e.predicate, alias, None)
            self.on_msg[e.name][MonitorState.ACTIVE].append(datum)


###############################################################################
# Requirement State Machine
###############################################################################


class RequirementBuilder(PatternBasedBuilder):
    def __init__(self, hpl_property):
        self.has_trigger_refs = False
        self.dependent_predicates = defaultdict(HplVacuousTruth)
        self.trigger_is_simple = hpl_property.pattern.trigger.is_simple_event
        self._behaviour = None
        event = hpl_property.pattern.behaviour
        if event.is_simple_event:
            self._behaviour = event.alias
            if event.alias:
                for e in hpl_property.pattern.trigger.simple_events():
                    if e.contains_reference(event.alias):
                        self.has_trigger_refs = True
                        break
        super().__init__(hpl_property, MonitorState.ACTIVE)

    @property
    def has_safe_state(self):
        return (self.timeout > 0 or self.reentrant_scope) and not self.has_trigger_refs

    def calc_pool_size(self, hpl_property):
        if not self.has_trigger_refs:
            if self.timeout > 0:
                return 1
            else:
                return 0
        return -1

    def add_terminator(self, event):
        # must be called before pattern events
        v = None if self.reentrant_scope else True
        for e in event.simple_events():
            alias = None
            if self._activator and e.contains_reference(self._activator):
                alias = self._activator
            datum = TerminatorEvent(e.predicate, alias, v)
            states = self.on_msg[e.name]
            states[MonitorState.ACTIVE].append(datum)
            if self.has_safe_state:
                states[MonitorState.SAFE].append(datum)

    def add_behaviour(self, event):
        for e in event.simple_events():
            # alias = None
            # if self._activator and e.contains_reference(self._activator):
            #    alias = self._activator
            # FIXME adding activator always to ensure dependent triggers have it
            datum = BehaviourEvent(e.predicate, self._activator, None)
            self.on_msg[e.name][MonitorState.ACTIVE].append(datum)

    def add_trigger(self, event):
        for e in event.simple_events():
            alias = None
            if self._activator and e.contains_reference(self._activator):
                alias = self._activator
            if self._behaviour:
                phi = e.predicate
                phi, psi = refactor_reference(phi, self._behaviour)
                if not psi.is_vacuous:
                    psi = replace_this_with_var(psi, '1')
                    psi = replace_var_with_this(psi, self._behaviour)
                    self.dependent_predicates[e.name] = psi
                datum = TriggerEvent(phi, alias)
            else:
                datum = TriggerEvent(e.predicate, alias)
            states = self.on_msg[e.name]
            states[MonitorState.ACTIVE].append(datum)
            if self.has_safe_state:
                states[MonitorState.SAFE].append(datum)


###############################################################################
# Response State Machine
###############################################################################


class ResponseBuilder(PatternBasedBuilder):
    def __init__(self, hpl_property):
        super().__init__(hpl_property, MonitorState.SAFE)

    def calc_pool_size(self, hpl_property):
        if self._trigger:
            for e in hpl_property.pattern.behaviour.simple_events():
                if e.contains_reference(self._trigger):
                    return -1
        # no alias, no refs
        return 1 if self.timeout >= 0 else 0

    def add_terminator(self, event):
        # must be called before pattern events
        for e in event.simple_events():
            alias = None
            if self._activator and e.contains_reference(self._activator):
                alias = self._activator
            states = self.on_msg[e.name]
            datum = TerminatorEvent(e.predicate, alias, False)
            states[MonitorState.ACTIVE].append(datum)
            if self.reentrant_scope:
                datum = TerminatorEvent(e.predicate, alias, None)
                states[MonitorState.SAFE].append(datum)
            else:
                datum = TerminatorEvent(e.predicate, alias, True)
                states[MonitorState.SAFE].append(datum)

    def add_behaviour(self, event):
        for e in event.simple_events():
            activator = None
            if self._activator and e.contains_reference(self._activator):
                activator = self._activator
            trigger = None
            if self._trigger and e.contains_reference(self._trigger):
                trigger = self._trigger
            datum = BehaviourEvent(e.predicate, activator, trigger)
            self.on_msg[e.name][MonitorState.ACTIVE].append(datum)

    def add_trigger(self, event):
        for e in event.simple_events():
            alias = None
            if self._activator and e.contains_reference(self._activator):
                alias = self._activator
            datum = TriggerEvent(e.predicate, alias)
            states = self.on_msg[e.name]
            if self.pool_size != 0:
                states[MonitorState.ACTIVE].append(datum)
            states[MonitorState.SAFE].append(datum)


###############################################################################
# Prevention State Machine
###############################################################################


class PreventionBuilder(PatternBasedBuilder):
    def __init__(self, hpl_property):
        super().__init__(hpl_property, MonitorState.SAFE)

    def calc_pool_size(self, hpl_property):
        if self._trigger:
            for e in hpl_property.pattern.behaviour.simple_events():
                if e.contains_reference(self._trigger):
                    return -1
        # no alias, no refs
        return 1  # if self.timeout >= 0 else 0

    def add_terminator(self, event):
        # must be called before pattern events
        for e in event.simple_events():
            alias = None
            if self._activator and e.contains_reference(self._activator):
                alias = self._activator
            if self.reentrant_scope:
                datum = TerminatorEvent(e.predicate, alias, None)
            else:
                datum = TerminatorEvent(e.predicate, alias, True)
            states = self.on_msg[e.name]
            states[MonitorState.ACTIVE].append(datum)
            states[MonitorState.SAFE].append(datum)

    def add_behaviour(self, event):
        for e in event.simple_events():
            activator = None
            if self._activator and e.contains_reference(self._activator):
                activator = self._activator
            trigger = None
            if self._trigger and e.contains_reference(self._trigger):
                trigger = self._trigger
            datum = BehaviourEvent(e.predicate, activator, trigger)
            self.on_msg[e.name][MonitorState.ACTIVE].append(datum)

    def add_trigger(self, event):
        for e in event.simple_events():
            alias = None
            if self._activator and e.contains_reference(self._activator):
                alias = self._activator
            datum = TriggerEvent(e.predicate, alias)
            states = self.on_msg[e.name]
            if self.pool_size != 0:
                states[MonitorState.ACTIVE].append(datum)
            states[MonitorState.SAFE].append(datum)



# SPDX-License-Identifier: MIT
# Copyright © 2023 André Santos

###############################################################################
# Imports
###############################################################################

from hplrv.monitors import MonitorState

from .common_data import *

###############################################################################
# Global Scope No Timeout
###############################################################################

def globally_some():
    text = 'globally: some b {len(xs) > 0}'
    traces = []
    # valid
    traces.append([])
    traces.append([ TimerEvent() ])
    traces.append([ SpamEvent('b', EMPTY_ARRAY) ])
    traces.append([
        SpamEvent('b', EMPTY_ARRAY),
        TimerEvent(),
        SpamEvent('b', EMPTY_ARRAY),
        BehaviourEvent(ARRAY_123, MonitorState.TRUE),
        SpamEvent('b', ARRAY_123),
    ])
    # invalid
    # none, lol
    return (text, traces)

###############################################################################
# Global Scope With Timeout
###############################################################################

def globally_some_within():
    text = 'globally: some b {exists i in xs: xs[@i] > 0} within 3 s'
    traces = []
    # valid
    traces.append([])
    traces.append([ TimerEvent() ])
    traces.append([ SpamEvent('b', EMPTY_ARRAY) ])
    traces.append([ BehaviourEvent(ARRAY_010, MonitorState.TRUE) ])
    traces.append([
        SpamEvent('b', Array(())),
        BehaviourEvent(ARRAY_010, MonitorState.TRUE),
        TimerEvent(),
        SpamEvent('b', ARRAY_010),
    ])
    # invalid
    traces.append([
        SpamEvent('b', ARRAY_000),
        TimerEvent(),
        SpamEvent('b', ARRAY_010, state=MonitorState.FALSE),
    ])
    traces.append([
        TimerEvent(),
        TimerEvent(),
        TimerEvent(state=MonitorState.FALSE),
    ])
    return (text, traces)

###############################################################################
# After Scope No Timeout
###############################################################################

def after_some():
    text = 'after p as P {xs[1] > 0}: some b {xs[1] > @P.xs[1]}'
    traces = []
    # valid
    traces.append([])
    traces.append([ TimerEvent() ])
    traces.append([ SpamEvent('b', ARRAY_010) ])
    traces.append([ ActivatorEvent(ARRAY_010, MonitorState.ACTIVE) ])
    traces.append([
        SpamEvent('b', ARRAY_123),
        ActivatorEvent(ARRAY_010, MonitorState.ACTIVE),
        TimerEvent(),
        SpamEvent('b', ARRAY_010),
        BehaviourEvent(ARRAY_123, MonitorState.TRUE),
        SpamEvent('b', ARRAY_123),
    ])
    # invalid
    # none, lol
    return (text, traces)

###############################################################################
# After Scope With Timeout
###############################################################################

def after_some_within():
    text = 'after p as P {xs[1] > 0}: some b {xs[1] > @P.xs[1]} within 3 s'
    traces = []
    # valid
    traces.append([])
    traces.append([ TimerEvent() ])
    traces.append([ SpamEvent('b', EMPTY_ARRAY) ])
    traces.append([ ActivatorEvent(ARRAY_010, MonitorState.ACTIVE) ])
    traces.append([
        ActivatorEvent(ARRAY_010, MonitorState.ACTIVE),
        BehaviourEvent(ARRAY_123, MonitorState.TRUE)
    ])
    traces.append([
        SpamEvent('b', ARRAY_010),
        TimerEvent(),
        ActivatorEvent(ARRAY_010, MonitorState.ACTIVE),
        SpamEvent('b', ARRAY_010),
        BehaviourEvent(ARRAY_123, MonitorState.TRUE),
        TimerEvent(),
        SpamEvent('b', ARRAY_123),
    ])
    # invalid
    traces.append([
        SpamEvent('b', ARRAY_010),
        ActivatorEvent(ARRAY_010, MonitorState.ACTIVE),
        TimerEvent(),
        SpamEvent('b', ARRAY_010),
        SpamEvent('b', ARRAY_123, state=MonitorState.FALSE),
        SpamEvent('b', ARRAY_123),
    ])
    return (text, traces)

###############################################################################
# Until Scope No Timeout
###############################################################################

def until_some():
    text = 'until q {forall i in xs: xs[@i] = 0}: some b {sum(xs) > 3}'
    traces = []
    # valid
    traces.append([])
    traces.append([ TimerEvent() ])
    traces.append([ SpamEvent('b', EMPTY_ARRAY) ])
    traces.append([ SpamEvent('q', ARRAY_010) ])
    traces.append([ BehaviourEvent(ARRAY_123, MonitorState.TRUE) ])
    traces.append([
        SpamEvent('b', ARRAY_000),
        TimerEvent(),
        SpamEvent('q', ARRAY_010),
    ])
    traces.append([
        BehaviourEvent(ARRAY_123, MonitorState.TRUE),
        SpamEvent('q', ARRAY_000),
    ])
    traces.append([
        SpamEvent('b', EMPTY_ARRAY),
        TimerEvent(),
        SpamEvent('b', ARRAY_111),
        SpamEvent('q', ARRAY_111),
        TimerEvent(),
        BehaviourEvent(ARRAY_123, MonitorState.TRUE),
        TimerEvent(),
        SpamEvent('b', ARRAY_123),
    ])
    # invalid
    traces.append([ TerminatorEvent(EMPTY_ARRAY, MonitorState.FALSE) ])
    traces.append([
        SpamEvent('b', ARRAY_111),
        SpamEvent('q', ARRAY_111),
        TimerEvent(),
        TerminatorEvent(ARRAY_000, MonitorState.FALSE),
        SpamEvent('b', ARRAY_123),
    ])
    return (text, traces)

###############################################################################
# Until Scope With Timeout
###############################################################################

def until_some_within():
    text = 'until q {forall i in xs: xs[@i] = 0}: some b {sum(xs) > 3} within 3 s'
    traces = []
    # valid
    traces.append([])
    traces.append([ TimerEvent() ])
    traces.append([ SpamEvent('b', EMPTY_ARRAY) ])
    traces.append([ SpamEvent('q', ARRAY_010) ])
    traces.append([ BehaviourEvent(ARRAY_123, MonitorState.TRUE) ])
    traces.append([
        BehaviourEvent(ARRAY_123, MonitorState.TRUE),
        SpamEvent('q', ARRAY_000),
    ])
    traces.append([
        SpamEvent('b', EMPTY_ARRAY),
        BehaviourEvent(ARRAY_123, MonitorState.TRUE),
        TimerEvent(),
        SpamEvent('b', ARRAY_111),
        SpamEvent('q', ARRAY_111),
        TimerEvent(),
        SpamEvent('b', ARRAY_123),
    ])
    # invalid
    traces.append([ TerminatorEvent(EMPTY_ARRAY, MonitorState.FALSE) ])
    traces.append([
        SpamEvent('b', ARRAY_000),
        TimerEvent(),
        SpamEvent('q', ARRAY_010, state=MonitorState.FALSE),
        SpamEvent('b', ARRAY_123),
    ])
    traces.append([
        TimerEvent(),
        TerminatorEvent(ARRAY_000, MonitorState.FALSE),
        SpamEvent('b', ARRAY_123),
    ])
    return (text, traces)

###############################################################################
# After-Until Scope No Timeout
###############################################################################

def after_until_some():
    text = 'after p as P {prod(xs) = 1} until q {max(xs) < min(@P.xs)}: some b {prod(xs) >= prod(@P.xs)}'
    traces = []
    # valid
    traces.append([])
    traces.append([ TimerEvent() ])
    traces.append([ SpamEvent('p', ARRAY_010) ])
    traces.append([ SpamEvent('q', EMPTY_ARRAY) ])
    traces.append([ SpamEvent('b', EMPTY_ARRAY) ])
    traces.append([ ActivatorEvent(EMPTY_ARRAY, MonitorState.ACTIVE) ])
    traces.append([
        SpamEvent('b', EMPTY_ARRAY),
        ActivatorEvent(ARRAY_111, MonitorState.ACTIVE),
        TimerEvent(),
        SpamEvent('b', ARRAY_010),
    ])
    traces.append([
        SpamEvent('p', ARRAY_010),
        SpamEvent('q', ARRAY_000),
        SpamEvent('b', ARRAY_111),
        ActivatorEvent(ARRAY_111, MonitorState.ACTIVE),
        SpamEvent('b', ARRAY_000),
        BehaviourEvent(ARRAY_123, MonitorState.SAFE),
        SpamEvent('q', ARRAY_123),
        TerminatorEvent(ARRAY_000, MonitorState.INACTIVE),
        SpamEvent('b', ARRAY_123),
        SpamEvent('q', ARRAY_000),
        ActivatorEvent(ARRAY_111, MonitorState.ACTIVE),
        SpamEvent('q', ARRAY_010),
        BehaviourEvent(ARRAY_123, MonitorState.SAFE),
        SpamEvent('b', ARRAY_123),
        TerminatorEvent(ARRAY_000, MonitorState.INACTIVE),
    ])
    # invalid
    traces.append([
        ActivatorEvent(ARRAY_111, MonitorState.ACTIVE),
        TerminatorEvent(ARRAY_000, MonitorState.FALSE)
    ])
    traces.append([
        SpamEvent('p', ARRAY_010),
        SpamEvent('q', ARRAY_000),
        SpamEvent('b', ARRAY_111),
        ActivatorEvent(ARRAY_111, MonitorState.ACTIVE),
        SpamEvent('b', ARRAY_000),
        SpamEvent('q', ARRAY_123),
        BehaviourEvent(ARRAY_123, MonitorState.SAFE),
        TerminatorEvent(ARRAY_000, MonitorState.INACTIVE),
        SpamEvent('b', ARRAY_123),
        SpamEvent('q', ARRAY_000),
        ActivatorEvent(ARRAY_111, MonitorState.ACTIVE),
        SpamEvent('b', ARRAY_010),
        SpamEvent('q', ARRAY_010),
        TerminatorEvent(ARRAY_000, MonitorState.FALSE),
    ])
    return (text, traces)

###############################################################################
# After-Until Scope With Timeout
###############################################################################

def after_until_some_within():
    text = 'after p as P {prod(xs) = 1} until q {max(xs) < min(@P.xs)}: some b {prod(xs) >= prod(@P.xs)} within 3 s'
    traces = []
    # valid
    traces.append([])
    traces.append([ TimerEvent() ])
    traces.append([ SpamEvent('p', ARRAY_010) ])
    traces.append([ SpamEvent('q', EMPTY_ARRAY) ])
    traces.append([ SpamEvent('b', EMPTY_ARRAY) ])
    traces.append([ ActivatorEvent(EMPTY_ARRAY, MonitorState.ACTIVE) ])
    traces.append([
        SpamEvent('b', EMPTY_ARRAY),
        ActivatorEvent(ARRAY_111, MonitorState.ACTIVE),
        TimerEvent(),
        SpamEvent('b', ARRAY_010),
    ])
    traces.append([
        SpamEvent('p', ARRAY_010),
        SpamEvent('q', ARRAY_000),
        SpamEvent('b', ARRAY_111),
        ActivatorEvent(ARRAY_111, MonitorState.ACTIVE),
        SpamEvent('b', ARRAY_000),
        BehaviourEvent(ARRAY_123, MonitorState.SAFE),
        SpamEvent('q', ARRAY_123),
        TerminatorEvent(ARRAY_000, MonitorState.INACTIVE),
        SpamEvent('b', ARRAY_123),
        SpamEvent('q', ARRAY_000),
        ActivatorEvent(ARRAY_111, MonitorState.ACTIVE),
        SpamEvent('q', ARRAY_010),
        BehaviourEvent(ARRAY_123, MonitorState.SAFE),
        SpamEvent('b', ARRAY_123),
        TerminatorEvent(ARRAY_000, MonitorState.INACTIVE),
    ])
    # invalid
    traces.append([
        ActivatorEvent(ARRAY_111, MonitorState.ACTIVE),
        TerminatorEvent(ARRAY_000, MonitorState.FALSE)
    ])
    traces.append([
        SpamEvent('p', ARRAY_010),
        SpamEvent('q', ARRAY_000),
        SpamEvent('b', ARRAY_111),
        ActivatorEvent(ARRAY_111, MonitorState.ACTIVE),
        SpamEvent('b', ARRAY_000),
        BehaviourEvent(ARRAY_123, MonitorState.SAFE),
        SpamEvent('q', ARRAY_123),
        TerminatorEvent(ARRAY_000, MonitorState.INACTIVE),
        SpamEvent('b', ARRAY_123),
        SpamEvent('q', ARRAY_000),
        ActivatorEvent(ARRAY_111, MonitorState.ACTIVE),
        SpamEvent('b', ARRAY_010),
        SpamEvent('q', ARRAY_010),
        SpamEvent('b', ARRAY_123, state=MonitorState.FALSE),
        SpamEvent('q', ARRAY_000),
    ])
    return (text, traces)



# SPDX-License-Identifier: MIT
# Copyright © 2023 André Santos

"""
Module that contains the command line program.

Why does this file exist, and why not put this in __main__?

  In some cases, it is possible to import `__main__.py` twice.
  This approach avoids that. Also see:
  https://click.palletsprojects.com/en/5.x/setuptools/#setuptools-integration

Some of the structure of this file came from this StackExchange question:
  https://softwareengineering.stackexchange.com/q/418600
"""

###############################################################################
# Imports
###############################################################################

from typing import Any, Dict, Final, List, Optional

import argparse
import sys
from traceback import print_exc

from hplrv import __version__ as current_version, gen, gui, play

###############################################################################
# Constants
###############################################################################

PROG: Final[str] = 'hpl-rv'
CMD_GEN: Final[str] = 'gen'
CMD_GUI: Final[str] = 'gui'
CMD_PLAY: Final[str] = 'play'

###############################################################################
# Argument Parsing
###############################################################################


def parse_arguments(argv: Optional[List[str]]) -> Dict[str, Any]:
    description = 'Tools to enable Runtime Verification from HPL properties.'
    parser = argparse.ArgumentParser(prog=PROG, description=description)

    parser.add_argument(
        '--version',
        action='version',
        version=f'{PROG} {current_version}',
        help='prints the program version',
    )

    parser.add_argument(
        'cmd',
        metavar='CMD',
        choices=[CMD_GEN, CMD_GUI, CMD_PLAY],
        help=f'a {PROG} command to run',
    )

    parser.add_argument(
        'args',
        metavar='ARG',
        nargs=argparse.REMAINDER,
        help=f'arguments for the {PROG} command',
    )

    args = parser.parse_args(args=argv)
    return vars(args)


###############################################################################
# Setup
###############################################################################


def load_configs(args: Dict[str, Any]) -> Dict[str, Any]:
    try:
        config: Dict[str, Any] = {}
        # with open(args['config_path'], 'r') as file_pointer:
        # yaml.safe_load(file_pointer)

        # arrange and check configs here

        return config
    except Exception as err:
        # log or raise errors
        print(err, file=sys.stderr)
        if str(err) == 'Really Bad':
            raise err

        # Optional: return some sane fallback defaults.
        sane_defaults: Dict[str, Any] = {}
        return sane_defaults


###############################################################################
# Entry Point
###############################################################################


def main(argv: Optional[List[str]] = None) -> int:
    args = parse_arguments(argv)

    try:
        # Load additional config files here, e.g., from a path given via args.
        # Alternatively, set sane defaults if configuration is missing.
        config = load_configs(args)
        cmd: str = args['cmd']

        if cmd == CMD_GEN:
            return gen.subprogram(args.get('args'), config)
        if cmd == CMD_GUI:
            return gui.subprogram(args.get('args'), config)
        if cmd == CMD_PLAY:
            return play.subprogram(args.get('args'), config)

    except KeyboardInterrupt:
        print('Aborted manually.', file=sys.stderr)
        return 1

    except Exception as err:
        print('An unhandled exception crashed the application!')
        print(err)
        print_exc()
        return 1

    return 0  # success



# SPDX-License-Identifier: MIT
# Copyright © 2023 André Santos

"""
Module that contains the 'replay' command line program and a utility
to replay traces of messages to test monitors.
"""

###############################################################################
# Imports
###############################################################################

from typing import Any, Dict, Final, List, Optional

import argparse
import asyncio
import importlib.util
import json
from pathlib import Path
from random import shuffle
import sys
from threading import Thread
from time import sleep

from hplrv.traces import Trace

###############################################################################
# Constants
###############################################################################

PROG_PLAY: Final[str] = 'hpl-rv play'


def noop(*args, **kwargs):
    pass


###############################################################################
# Setup Functions
###############################################################################


def import_generated_monitors(module_name: str, file_path: Path):
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)
    return module


def import_trace_from_json_file(file_path: Path) -> Trace:
    text: str = file_path.read_text(encoding='utf8')
    data: List[Dict[str, Any]] = json.loads(text)
    return Trace.from_list_of_dict(data)


def print_monitor_success(monitor, timestamp, witness):
    print(f'> Success @{timestamp}s')
    print(f'  [HPL]: {monitor.HPL_PROPERTY}')
    print(f'  [witness]: {witness}')


def print_monitor_failure(monitor, timestamp, witness):
    print(f'> Failure @{timestamp}s')
    print(f'  [HPL]: {monitor.HPL_PROPERTY}')
    print(f'  [witness]: {witness}')


###############################################################################
# Public Interface
###############################################################################


def trace_replay(monitor, trace: Trace, freq: float, shutdown: bool = True) -> float:
    now: float = 0.0
    next: float = freq
    monitor.launch(now)

    for event in trace.events:
        t = event.timestamp

        # wait until the next timestamp with messages
        while next < t:
            sleep(next - now)
            now = next
            next += freq
            monitor.on_timer(now)
        assert next >= t, f'{next} < {t}'
        if next > t:
            # wait the remaining time until message event
            sleep(t - now)
        else:
            next += freq
        now = t

        # dispatch simultaneous messages in a random order
        msgs = list(event.messages)
        shuffle(msgs)
        for msg in msgs:
            topic = msg.topic.replace('/', '_')
            cb = getattr(monitor, f'on_msg_{topic}', noop)
            cb(msg.data, now)

    # wait until the next timer event
    sleep(next - now)
    monitor.on_timer(now)
    # end of trace
    if shutdown:
        monitor.shutdown(now)
    return now


###############################################################################
# Entry Point
###############################################################################


def subprogram(
    argv: Optional[List[str]],
    _settings: Optional[Dict[str, Any]] = None,
) -> int:
    args = parse_arguments(argv)
    return run(args, _settings or {})


def run(args: Dict[str, Any], _settings: Dict[str, Any]) -> int:
    file_path: Optional[Path] = args['module'].resolve(strict=True)
    lib = import_generated_monitors(file_path.stem, file_path)
    file_path = args.get('data')
    if file_path is None:
        return 0
    trace: Trace = import_trace_from_json_file(file_path)
    freq: float = args.get('frequency', 1.0)
    man = lib.HplMonitorManager()
    man.on_monitor_success = print_monitor_success
    man.on_monitor_failure = print_monitor_failure
    man.live_server.host = args['host']
    man.live_server.port = args['port']
    thread: Thread = man.live_server.start_thread()
    timestamp = trace_replay(man, trace, freq, shutdown=False)
    print('End of trace.')
    print('Press Ctrl+C to exit.')
    try:
        while True:
            sleep(1.0)
    except KeyboardInterrupt:
        pass
    man.shutdown(timestamp)
    thread.join(10.0)
    return 0


###############################################################################
# Argument Parsing
###############################################################################


def parse_arguments(argv: Optional[List[str]]) -> Dict[str, Any]:
    description = 'Utility to run monitors and replay traces.'
    parser = argparse.ArgumentParser(prog=PROG_PLAY, description=description)

    parser.add_argument(
        'module',
        type=Path,
        help='path to a module with generated monitors'
    )

    parser.add_argument(
        '-d',
        '--data',
        type=Path,
        help=f'path to a data file containing a message trace',
    )

    parser.add_argument(
        '--host',
        default='127.0.0.1',
        help='address of the live monitoring server (default: 127.0.0.1)'
    )

    parser.add_argument(
        '-p',
        '--port',
        type=int,
        default=4242,
        help=f'port of the live monitoring server (default: 4242)',
    )

    parser.add_argument(
        '-f',
        '--frequency',
        type=float,
        default=1.0,
        help=f'timestamp increment between trace events (default: 1)',
    )

    args = parser.parse_args(args=argv)
    return vars(args)



# SPDX-License-Identifier: MIT
# Copyright © 2023 André Santos

###############################################################################
# Imports
###############################################################################

from typing import Any, Iterable, List, Mapping, Optional, Tuple

from bisect import bisect, bisect_left
from types import SimpleNamespace

from attrs import field, frozen

###############################################################################
# Data Structures
###############################################################################


@frozen
class Message:
    topic: str
    data: Any = field(order=False)

    @classmethod
    def from_dict(cls, data: Mapping[str, Any]) -> 'Message':
        return cls(data['topic'], SimpleNamespace(**data.get('data', {})))


@frozen(order=True)
class TraceEvent:
    timestamp: float
    messages: Tuple[Message] = field(factory=tuple, order=False, converter=tuple)

    @classmethod
    def from_dict(cls, data: Mapping[str, Any]) -> 'TraceEvent':
        messages = map(Message.from_dict, data.get('messages', []))
        return cls(data['timestamp'], messages)

    def merge(self, other: 'TraceEvent') -> 'TraceEvent':
        if self.timestamp != other.timestamp:
            raise ValueError(f'timestamps differ: {self.timestamp} != {other.timestamp}')
        messages = self.messages + other.messages
        return TraceEvent(self.timestamp, messages)


@frozen
class Trace:
    events: Tuple[TraceEvent] = field(factory=tuple, converter=tuple)

    @events.validator
    def _ensure_sorted_timestamps(self, _attr: Any, events: Iterable[TraceEvent]) -> None:
        if not is_sorted(events):
            raise ValueError(f'unsorted events: {events}')

    @classmethod
    def from_unsorted_events(cls, unsorted_events: Iterable[TraceEvent]) -> 'Trace':
        # ensure that all duplicate timestamps are merged
        events = []
        for event in unsorted_events:
            insort_event(events, event)
        return cls(events)

    @classmethod
    def from_list_of_dict(cls, data: Iterable[Mapping[str, Any]]) -> 'Trace':
        events = map(TraceEvent.from_dict, data)
        return Trace(events)

    def add(self, event: TraceEvent) -> 'Trace':
        events = list(self.events)
        insort_event(events, event)
        return Trace(events)

    def merge(self, other: 'Trace') -> 'Trace':
        events = list(self.events)
        for event in other.events:
            insort_event(events, event)
        return Trace(events)

    def previous_timestamp(self, timestamp: float) -> Optional[float]:
        # returns the timestamp of the first event
        # that comes before (<) the given timestamp, or None
        i = bisect_left(self.events, timestamp, key=get_timestamp)
        return None if i <= 0 else self.events[i-1].timestamp

    def next_timestamp(self, timestamp: float) -> Optional[float]:
        # returns the timestamp of the first event
        # that comes after (>) the given timestamp, or None
        i = bisect(self.events, timestamp, key=get_timestamp)
        return None if i >= len(self.events) else self.events[i].timestamp


###############################################################################
# Helper Functions
###############################################################################


def is_sorted(items: Iterable[Any]) -> bool:
    return all(items[i] <= items[i+1] for i in range(len(items) - 1))


def insort_event(events: List[TraceEvent], event: TraceEvent) -> None:
    i = bisect(events, event)
    if i > 0 and events[i-1].timestamp == event.timestamp:
        events[i-1] = events[i-1].merge(event)
    else:
        events.insert(i, event)


def get_timestamp(event: TraceEvent) -> float:
    return event.timestamp



# SPDX-License-Identifier: MIT
# Copyright © 2023 André Santos

###############################################################################
# Imports
###############################################################################

from typing import Any, Final, Iterable, Optional

from attrs import frozen

from hplrv.monitors import EventType, MonitorState

###############################################################################
# Data Structures
###############################################################################


@frozen
class MsgRecord:
    topic: str
    timestamp: float
    msg: Any


@frozen
class ActivatorEvent:
    msg: Any
    state: MonitorState
    topic: str = 'p'

    @property
    def event(self) -> EventType:
        return EventType.ACTIVATOR


@frozen
class TerminatorEvent:
    msg: Any
    state: MonitorState
    topic: str = 'q'

    @property
    def event(self) -> EventType:
        return EventType.TERMINATOR


@frozen
class BehaviourEvent:
    msg: Any
    state: MonitorState
    topic: str = 'b'

    @property
    def event(self) -> EventType:
        return EventType.BEHAVIOUR


@frozen
class TriggerEvent:
    msg: Any
    state: MonitorState
    topic: str = 'a'

    @property
    def event(self) -> EventType:
        return EventType.TRIGGER


@frozen
class SpamEvent:
    topic: str
    msg: Any
    state: Optional[MonitorState] = None

    @property
    def event(self) -> EventType:
        return EventType.SPAM


@frozen
class TimerEvent:
    state: Optional[MonitorState] = None
    drops: int = 0

    @property
    def event(self) -> EventType:
        return EventType.TIMER


@frozen
class Point2D:
    x: float = 0
    y: float = 0


@frozen
class Array:
    xs: Iterable[int]


EMPTY_ARRAY: Final[Array] = Array(())
ARRAY_000: Final[Array] = Array((0, 0, 0))
ARRAY_010: Final[Array] = Array((0, 1, 0))
ARRAY_111: Final[Array] = Array((1, 1, 1))
ARRAY_123: Final[Array] = Array((1, 2, 3))



# SPDX-License-Identifier: MIT
# Copyright © 2023 André Santos

"""
Module that contains the 'gui' command line program and the
dashboard server to manage runtime monitors.
"""

###############################################################################
# Imports
###############################################################################

from typing import Any, Callable, Dict, Final, List, Optional, Set, Tuple, Type

import argparse
import json
from pathlib import Path
import pkg_resources

from attrs import define, field, frozen
from bottle import Bottle, request, response, run as serve_forever, static_file
from bottle.ext.websocket import GeventWebSocketServer, websocket
import gevent
from gevent.queue import Queue
from gevent.socket import create_connection, socket

# from hpl.ast import HplProperty, HplSpecification
# from hpl.parser import property_parser, specification_parser

###############################################################################
# Constants
###############################################################################

PROG_GUI: Final[str] = 'hpl-rv gui'

CLIENT_ROOT: Final[str] = str(
    Path(pkg_resources.resource_filename('hplrv', 'dashboard/index.html'))
    .resolve(strict=True)
    .parent
)

WS_UPDATE_CB_TYPE: Final[Type] = Callable[[Dict[str, Any]], None]

COMPACT: Final[Tuple[str, str]] = (',', ':')


def noop(*args, **kwargs):
    pass


###############################################################################
# Live Monitoring Client/Server
###############################################################################


@define
class SocketStreamReader:
    # taken from
    # https://stackoverflow.com/a/65637828

    _sock: socket
    _recv_buffer: bytearray = field(factory=bytearray, init=False, eq=False)

    def read(self, num_bytes: int = -1) -> bytes:
        raise NotImplementedError

    def readline(self) -> str:
        return self.readuntil(b'\n').decode('utf8')

    def readuntil(self, separator: bytes = b'\n') -> bytes:
        if len(separator) != 1:
            raise ValueError('Only separators of length 1 are supported.')

        chunk = bytearray(4096)
        start = 0
        buf = bytearray(len(self._recv_buffer))
        bytes_read = self._recv_into(memoryview(buf))
        assert bytes_read == len(buf)

        while True:
            idx = buf.find(separator, start)
            if idx != -1:
                break

            start = len(self._recv_buffer)
            bytes_read = self._recv_into(memoryview(chunk))
            buf += memoryview(chunk)[:bytes_read]

        result = bytes(buf[: idx + 1])
        self._recv_buffer = b''.join(
            (memoryview(buf)[idx + 1 :], self._recv_buffer)
        )
        return result

    def _recv_into(self, view: memoryview) -> int:
        bytes_read = min(len(view), len(self._recv_buffer))
        view[:bytes_read] = self._recv_buffer[:bytes_read]
        self._recv_buffer = self._recv_buffer[bytes_read:]
        if bytes_read == len(view):
            return bytes_read
        bytes_read += self._sock.recv_into(view[bytes_read:])
        return bytes_read


@frozen
class LiveMonitoringServer:
    """
    A gevent greenlet to handle live monitoring status updates.
    If handling multiple multiple live monitoring sources, it is
    preferred to use a shared `updates` queue to make waiting easier.
    """

    host: str
    port: int
    on_update: WS_UPDATE_CB_TYPE = field(default=noop, eq=False, order=False)
    monitors: List[Dict[str, Any]] = field(factory=list, init=False, eq=False, order=False)
    _socket: Optional[socket] = field(default=None, init=False, eq=False, order=False)

    @property
    def is_connected(self) -> bool:
        return self._socket is not None

    def connect(self):
        if self._socket is None:
            s = create_connection((self.host, self.port))
            object.__setattr__(self, "_socket", s)

    def run(self):
        try:
            with self._socket:
                reader = SocketStreamReader(self._socket)

                # read initial monitor report
                data = reader.readline()
                if not data:
                    return
                self.monitors.clear()
                self.monitors.extend(json.loads(data))
                self.push_status()

                # read status updates
                data = reader.readline()
                while data:
                    verdict = json.loads(data)
                    self.push_verdict(verdict)
                    data = reader.readline()
        finally:
            object.__setattr__(self, "_socket", None)

    def push_status(self):
        for i in range(len(self.monitors)):
            monitor = self.monitors[i]
            self.on_update({
                'server': f'{self.host}:{self.port}',
                'id': i,
                'monitor': monitor,
            })

    def push_verdict(self, verdict: Dict[str, Any]):
        i = verdict['monitor']
        monitor = self.monitors[i]
        monitor['verdict'] = verdict['value']
        monitor['witness'] = verdict['witness']
        self.on_update({
            'server': f'{self.host}:{self.port}',
            'id': i,
            'monitor': monitor,
        })

    def asdict(self) -> Dict[str, Any]:
        return {
            'host': self.host,
            'port': self.port,
        }


@frozen
class LiveMonitoringClient:
    websocket: Any
    updates: Queue = field(factory=Queue, init=False, eq=False, order=False)

    def send_initial_server_status(self, host: str, port: int, monitors: List[Dict[str, Any]]):
        server: str = f'{host}:{port}'
        for i in range(len(monitors)):
            monitor = monitors[i]
            data = {
                'server': server,
                'id': i,
                'monitor': monitor,
            }
            self.send_monitor_status(data)

    def get_and_send_update(self):
        status = self.updates.get()
        self.send_monitor_status(status)

    def send_monitor_status(self, status: Dict[str, Any]):
        self.websocket.send(json.dumps(status, separators=COMPACT))


###############################################################################
# Public Interface
###############################################################################

# Important note regarding gevent-based servers:
# "... the servers spawn one greenlet per connection (not per request)"

# Normal HTTP requests create a new connection as per the WSGI standard.
# The websocket request is a new connection (of type websocket), so we have
# (h + w) connections going on at any given time (h = HTTP, w = WebSocket),
# meaning that we also have (h + w) greenlets.

# Greenlets:
# - main server loop?
# - one per HTTP request (automatic cleanup)
# - one per websocket (loop)
# - one per live monitoring server socket (loop)


@frozen
class MonitorServer:
    app: Bottle
    servers: Set[LiveMonitoringServer] = field(factory=set, init=False, eq=False, order=False)
    clients: Set[LiveMonitoringClient] = field(factory=set, init=False, eq=False, order=False)

    def __attrs_post_init__(self):
        self.app.get('/')(self.index)
        self.app.get('/<filename:path>')(self.send_static_file)
        self.app.post('/live')(self.connect_to_live_server)
        self.app.get('/ws', apply=[websocket])(self.live_updates)

    def index(self):
        return static_file('index.html', root=CLIENT_ROOT, mimetype='text/html')

    def send_static_file(self, filename):
        return static_file(filename, root=CLIENT_ROOT)

    def live_updates(self, ws):
        client = LiveMonitoringClient(ws)
        self.clients.add(client)
        try:
            # send initial status reports
            for server in self.servers:
                client.send_initial_server_status(server.host, server.port, server.monitors)
            while True:
                client.get_and_send_update()
        finally:
            self.clients.remove(client)

    def connect_to_live_server(self):
        # host: str = request.forms.get('host')
        # port: int = int(request.forms.get('port'))
        host: str = request.json.get('host')
        port: int = request.json.get('port')
        server = LiveMonitoringServer(host, port, on_update=self._on_live_server_update)
        if server in self.servers:
            return { 'servers': [s.asdict() for s in self.servers] }
        try:
            server.connect()
        except (OSError, ConnectionRefusedError) as e:
            # abort(502, 'Unable to connect to live monitoring server.')
            response.status = 502  # Bad Gateway
            return { 'error': repr(e) }
        else:
            self.servers.add(server)
            gevent.spawn(self._handle_live_server, server)
            return { 'servers': [s.asdict() for s in self.servers] }

    def _handle_live_server(self, server: LiveMonitoringServer):
        try:
            server.run()
        finally:
            self.servers.remove(server)

    def _on_live_server_update(self, update):
        # broadcast to all clients
        for client in self.clients:
            client.updates.put(update)


###############################################################################
# Entry Point
###############################################################################


def subprogram(
    argv: Optional[List[str]],
    _settings: Optional[Dict[str, Any]] = None,
) -> int:
    args = parse_arguments(argv)
    return run(args, _settings or {})


def run(args: Dict[str, Any], _settings: Dict[str, Any]) -> int:
    # TODO use gevent.kill(greenlet) or gevent.killall(greenlets, block=True, timeout=None)
    # with KeyboardInterrupt to kill all ongoing greenlets.
    server = MonitorServer(Bottle())
    serve_forever(app=server.app, host=args['host'], port=args['port'], server=GeventWebSocketServer)
    return 0


###############################################################################
# Argument Parsing
###############################################################################


def parse_arguments(argv: Optional[List[str]]) -> Dict[str, Any]:
    description = 'Dashboard to manage runtime monitors.'
    parser = argparse.ArgumentParser(prog=PROG_GUI, description=description)

    parser.add_argument(
        '--host',
        default='127.0.0.1',
        help='address of the HTTP server (default: 127.0.0.1)'
    )

    parser.add_argument(
        '-p',
        '--port',
        type=int,
        default=8080,
        help=f'port of the HTTP server (default: 8080)',
    )

    args = parser.parse_args(args=argv)
    return vars(args)



# SPDX-License-Identifier: MIT
# Copyright © 2023 André Santos

###############################################################################
# Imports
###############################################################################

from collections import deque       # needed by PropertyMonitor
from threading import Lock          # needed by PropertyMonitor

from hpl.parser import property_parser

from hplrv.gen import MonitorGenerator
from hplrv.monitors import EventType, MonitorState

from .common_data import *
from .absence_traces import *
from .existence_traces import *
from .precedence_traces import *
from .response_traces import *
from .prevention_traces import *

###############################################################################
# Test Case Generation
###############################################################################

def absence_properties():
    yield globally_no()
    yield globally_no_within()
    yield after_no()
    yield after_no_within()
    yield until_no()
    yield until_no_within()
    yield after_until_no()
    yield after_until_no_within()

def existence_properties():
    yield globally_some()
    yield globally_some_within()
    yield after_some()
    yield after_some_within()
    yield until_some()
    yield until_some_within()
    yield after_until_some()
    yield after_until_some_within()

def precedence_properties():
    yield globally_requires()
    yield globally_requires_ref()
    yield globally_requires_within()
    yield globally_requires_ref_within()
    yield after_requires()
    yield after_requires_ref()
    yield after_requires_within()
    yield after_requires_ref_within()
    yield until_requires()
    yield until_requires_ref()
    yield until_requires_within()
    yield until_requires_ref_within()
    yield after_until_requires()
    yield after_until_requires_ref()
    yield after_until_requires_within()
    yield after_until_requires_ref_within()

def response_properties():
    yield globally_causes()
    yield globally_causes_ref()
    yield globally_causes_within()
    yield globally_causes_ref_within()
    yield after_causes()
    yield after_causes_ref_within()
    yield until_causes()
    yield until_causes_ref()
    yield until_causes_within()
    yield until_causes_ref_within()
    yield after_until_causes()
    yield after_until_causes_ref()
    yield after_until_causes_within()
    yield after_until_causes_ref_within()

def prevention_properties():
    yield globally_forbids()
    yield globally_forbids_ref()
    yield globally_forbids_within()
    yield globally_forbids_ref_within()
    yield after_forbids()
    yield after_forbids_ref()
    yield after_forbids_within()
    yield after_forbids_ref_within()
    yield until_forbids()
    yield until_forbids_ref()
    yield until_forbids_within()
    yield until_forbids_ref_within()
    yield after_until_forbids()
    yield after_until_forbids_ref()
    yield after_until_forbids_within()
    yield after_until_forbids_ref_within()

def all_types_of_property():
    # example = text, traces
    for example in absence_properties():
        yield example
    for example in existence_properties():
        yield example
    for example in precedence_properties():
        yield example
    for example in response_properties():
        yield example
    for example in prevention_properties():
        yield example


###############################################################################
# Test Loop
###############################################################################


def state_name(s: MonitorState):
    if s is None:
        return '(none)'
    try:
        return s.name
    except AttributeError:
        return 'STATE {}'.format(s)


def pretty_trace(trace):
    s = []
    t = 0
    for e in trace:
        t += 1
        etype = e.event
        if etype == EventType.TIMER:
            s.append('@ {}: (Timer, -{}) -> {}'.format(
                t, e.drops, state_name(e.state)))
        elif etype == EventType.ACTIVATOR:
            s.append("@ {}: (Activator) '{}' {} -> {}".format(
                t, e.topic, e.msg, state_name(e.state)))
        elif etype == EventType.TERMINATOR:
            s.append("@ {}: (Terminator) '{}' {} -> {}".format(
                t, e.topic, e.msg, state_name(e.state)))
        elif etype == EventType.BEHAVIOUR:
            s.append("@ {}: (Behaviour) '{}' {} -> {}".format(
                t, e.topic, e.msg, state_name(e.state)))
        elif etype == EventType.TRIGGER:
            s.append("@ {}: (Trigger) '{}' {} -> {}".format(
                t, e.topic, e.msg, state_name(e.state)))
        else:
            s.append("@ {}: (Spam) '{}' {}".format(
                t, e.topic, e.msg))
    return "\n".join(s)

def pretty_monitor(m):
    return "\n".join((
        'm._state = {}'.format(m._state),
        'm.time_launch = {}'.format(m.time_launch),
        'm.time_shutdown = {}'.format(m.time_shutdown),
        'm.time_state = {}'.format(m.time_state),
        'm.witness = {}'.format(m.witness),
        'm._pool = {}'.format(getattr(m, '_pool', None)),
    ))


def prod(iterable):
    x = 1
    for y in iterable:
        x = x * y
        if x == 0:
            return 0
    return x


class TestMonitorClasses:
    def test_examples(self):
        self._reset()
        n = 0
        p = property_parser()
        r = MonitorGenerator(lang='py')
        for text, traces in all_types_of_property():
            hp = p.parse(text)
            self.pool_decay = ((hp.pattern.is_requirement
                    or hp.pattern.is_response
                    or hp.pattern.is_prevention)
                and hp.pattern.has_max_time)
            py = r.monitor_class(hp)
            m = self._make_monitor(py)
            for trace in traces:
                n += 1
                self.hpl_string = text
                self._set_trace_string(trace, n)
                self._launch(hp, m)
                time = 0
                for event in trace:
                    time += 1
                    self._dispatch(m, event, time)
                self._shutdown(m)
                self._reset()
        print('Tested {} examples.'.format(n))

    def _reset(self):
        self.debug_string = ''
        self.trace_string = ''
        self.hpl_string = ''
        self.entered_scope = []
        self.exited_scope = []
        self.found_success = []
        self.found_failure = []

    def _make_monitor(self, py):
        loc = dict(locals())
        exec(py, globals(), loc)
        cls = loc['PropertyMonitor']
        m = cls()
        m.on_enter_scope = self._on_enter
        m.on_exit_scope = self._on_exit
        m.on_success = self._on_success
        m.on_violation = self._on_failure
        self._update_debug_string(m, -1)
        assert m._state == MonitorState.OFF, self.debug_string
        assert m.verdict is None, self.debug_string
        assert not m.witness, self.debug_string
        assert m.time_launch < 0, self.debug_string
        assert m.time_state < 0, self.debug_string
        assert m.time_shutdown < 0, self.debug_string
        return m

    def _launch(self, hp, m):
        m.on_launch(0)
        self._update_debug_string(m, 0)
        if hp.scope.is_global or (hp.scope.is_until and not hp.scope.is_after):
            assert len(self.entered_scope) == 1, self.debug_string
            assert self.entered_scope[0] == 0, self.debug_string
        else:
            assert not self.entered_scope, self.debug_string
        assert not self.exited_scope, self.debug_string
        assert not self.found_success, self.debug_string
        assert not self.found_failure, self.debug_string
        assert m._state != MonitorState.OFF, self.debug_string
        assert m.verdict is None, self.debug_string
        assert not m.witness, self.debug_string
        assert m.time_launch == 0, self.debug_string
        assert m.time_state == 0, self.debug_string
        assert m.time_shutdown < 0, self.debug_string
        try:
            m.on_launch(0)
            self._update_debug_string(m, 0)
            assert False, self.debug_string
        except RuntimeError:
            pass # expected

    def _dispatch(self, m, event, time):
        etype = event.event
        if etype == EventType.TIMER:
            self._dispatch_timer(m, event, time)
        elif etype == EventType.ACTIVATOR:
            self._dispatch_activator(m, event, time)
        elif etype == EventType.TERMINATOR:
            self._dispatch_terminator(m, event, time)
        elif etype == EventType.BEHAVIOUR:
            self._dispatch_behaviour(m, event, time)
        elif etype == EventType.TRIGGER:
            self._dispatch_trigger(m, event, time)
        else:
            self._dispatch_spam(m, event, time)

    def _dispatch_activator(self, m, event, time):
        n = len(self.entered_scope)
        assert m._state == MonitorState.INACTIVE, self.debug_string
        consumed = self._dispatch_msg(m, event.topic, event.msg, time)
        self._update_debug_string(m, time)
        assert consumed, self.debug_string
        assert len(self.entered_scope) == n + 1, self.debug_string
        assert self.entered_scope[-1] == time, self.debug_string
        assert len(self.exited_scope) == n, self.debug_string
        assert not self.found_success, self.debug_string
        assert not self.found_failure, self.debug_string
        assert m._state == event.state, self.debug_string
        assert m.verdict is None, self.debug_string
        assert len(m.witness) == 1, self.debug_string
        assert m.witness[-1].topic == event.topic, self.debug_string
        assert m.witness[-1].timestamp == time, self.debug_string
        assert m.witness[-1].msg == event.msg, self.debug_string
        assert m.time_state == time, self.debug_string

    def _dispatch_terminator(self, m, event, time):
        n = len(self.exited_scope)
        assert m._state > MonitorState.INACTIVE, self.debug_string
        consumed = self._dispatch_msg(m, event.topic, event.msg, time)
        self._update_debug_string(m, time)
        assert consumed, self.debug_string
        assert len(self.exited_scope) == n + 1, self.debug_string
        assert self.exited_scope[-1] == time, self.debug_string
        assert len(self.entered_scope) == n + 1, self.debug_string
        self._check_verdict(m, event, time)
        if event.state == MonitorState.INACTIVE:
            assert not m.witness, self.debug_string
            assert not getattr(m, '_pool', None), self.debug_string

    def _dispatch_behaviour(self, m, event, time):
        a = len(self.entered_scope)
        b = len(self.exited_scope)
        s = m._state
        assert m._state == MonitorState.ACTIVE or m._state == MonitorState.SAFE, \
               self.debug_string
        consumed = self._dispatch_msg(m, event.topic, event.msg, time)
        self._update_debug_string(m, time)
        assert consumed, self.debug_string
        assert len(self.entered_scope) == a, self.debug_string
        assert len(self.exited_scope) == b, self.debug_string
        if event.state is None:
            assert m._state == s, self.debug_string
        self._check_verdict(m, event, time)

    def _dispatch_trigger(self, m, event, time):
        a = len(self.entered_scope)
        b = len(self.exited_scope)
        k = -1 if not hasattr(m, '_pool') else len(m._pool)
        s = m._state
        consumed = self._dispatch_msg(m, event.topic, event.msg, time)
        self._update_debug_string(m, time)
        assert consumed, self.debug_string
        assert len(self.entered_scope) == a, self.debug_string
        assert len(self.exited_scope) == b, self.debug_string
        if event.state is None:
            assert m._state == s, self.debug_string
        self._check_verdict(m, event, time)
        if k >= 0:
            assert len(m._pool) >= k, self.debug_string
            assert m._pool[-1].topic == event.topic, self.debug_string
            assert m._pool[-1].timestamp == time, self.debug_string
            assert m._pool[-1].msg == event.msg, self.debug_string

    def _dispatch_spam(self, m, event, time):
        a = len(self.entered_scope)
        b = len(self.exited_scope)
        k = len(getattr(m, '_pool', ()))
        n = len(m.witness)
        s = m._state
        t = m.time_state
        consumed = self._dispatch_msg(m, event.topic, event.msg, time)
        self._update_debug_string(m, time)
        assert not consumed, self.debug_string
        assert len(self.entered_scope) == a, self.debug_string
        assert len(self.exited_scope) == b, self.debug_string
        if self.pool_decay:
            assert len(getattr(m, '_pool', ())) <= k, self.debug_string
        else:
            assert len(getattr(m, '_pool', ())) == k, self.debug_string
        self._check_automatic_transition(m, event.state, time, s, t)

    def _dispatch_msg(self, m, topic, msg, time):
        cb = getattr(m, 'on_msg_' + topic)
        return cb(msg, time)

    def _dispatch_timer(self, m, event, time):
        a = len(self.entered_scope)
        b = len(self.exited_scope)
        k = len(getattr(m, '_pool', ()))
        s = m._state
        t = m.time_state
        m.on_timer(time)
        self._update_debug_string(m, time)
        assert len(self.entered_scope) == a, self.debug_string
        assert len(self.exited_scope) == b, self.debug_string
        assert len(getattr(m, '_pool', ())) <= k, self.debug_string
        self._check_automatic_transition(m, event.state, time, s, t)

    def _shutdown(self, m):
        m.on_shutdown(1000)
        self._update_debug_string(m, 1000)
        assert m._state == MonitorState.OFF, self.debug_string
        assert m.time_launch == 0, self.debug_string
        assert m.time_state >= 0, self.debug_string
        assert m.time_shutdown == 1000, self.debug_string
        try:
            m.on_shutdown(2000)
            self._update_debug_string(m, 2000)
            assert False, self.debug_string
        except RuntimeError:
            pass # expected

    def _check_verdict(self, m, event, time):
        if event.state is not None:
            assert m._state == event.state, self.debug_string
        if event.state == MonitorState.TRUE:
            assert len(self.found_success) == 1, self.debug_string
            assert self.found_success[0][0] == time, self.debug_string
            assert self.found_success[0][1] == m.witness, self.debug_string
            assert m.verdict is True, self.debug_string
            assert len(m.witness) >= 1, self.debug_string
            assert m.witness[-1].topic == event.topic, self.debug_string
            assert m.witness[-1].timestamp == time, self.debug_string
            assert m.witness[-1].msg == event.msg, self.debug_string
            assert m.time_state == time, self.debug_string
        elif event.state == MonitorState.FALSE:
            assert len(self.found_failure) == 1, self.debug_string
            assert self.found_failure[0][0] == time, self.debug_string
            assert self.found_failure[0][1] == m.witness, self.debug_string
            assert m.verdict is False, self.debug_string
            assert len(m.witness) >= 1, self.debug_string
            assert m.witness[-1].topic == event.topic, self.debug_string
            assert m.witness[-1].timestamp == time, self.debug_string
            assert m.witness[-1].msg == event.msg, self.debug_string
            assert m.time_state == time, self.debug_string
        else:
            assert not self.found_success, self.debug_string
            assert not self.found_failure, self.debug_string
            assert m.verdict is None, self.debug_string

    def _check_automatic_transition(self, m, s2, t2, s1, t1):
        if s2 == MonitorState.TRUE:
            assert len(self.found_success) == 1, self.debug_string
            assert self.found_success[0][0] == t2, self.debug_string
            assert self.found_success[0][1] == m.witness, self.debug_string
            assert m.verdict is True, self.debug_string
            assert m.time_state == t2, self.debug_string
            assert m._state == s2, self.debug_string
        elif s2 == MonitorState.FALSE:
            assert len(self.found_failure) == 1, self.debug_string
            assert self.found_failure[0][0] == t2, self.debug_string
            assert self.found_failure[0][1] == m.witness, self.debug_string
            assert m.verdict is False, self.debug_string
            assert m.time_state == t2, self.debug_string
            assert m._state == s2, self.debug_string
        elif s2 is not None:
            assert not self.found_success, self.debug_string
            assert not self.found_failure, self.debug_string
            assert m.verdict is None, self.debug_string
            assert m._state == s2, self.debug_string
        else:
            assert m._state == s1, self.debug_string
            assert m.time_state == t1, self.debug_string

    def _on_enter(self, stamp):
        self.entered_scope.append(stamp)

    def _on_exit(self, stamp):
        self.exited_scope.append(stamp)

    def _on_success(self, stamp, witness):
        self.found_success.append((stamp, witness))

    def _on_failure(self, stamp, witness):
        self.found_failure.append((stamp, witness))

    def _set_trace_string(self, trace, n):
        self.trace_string = f'[Example #{n}]:\n{pretty_trace(trace)}'

    def _update_debug_string(self, m, time):
        self.debug_string = (
            'failed for the following test'
            f'\n  [HPL]: {self.hpl_string}'
            f'\n  {self.trace_string}'
            f'\n  [Timestamp]: {time}'
            '\n  [Monitor]:'
            f'\n{pretty_monitor(m)}'
        )



# SPDX-License-Identifier: MIT
# Copyright © 2023 André Santos

###############################################################################
# Imports
###############################################################################

from hplrv.monitors import MonitorState

from .common_data import *

###############################################################################
# Global Scope No Timeout
###############################################################################

def globally_no():
    text = 'globally: no b {x > 0}'
    traces = []
    # valid
    traces.append([])
    traces.append([ TimerEvent() ])
    traces.append([ SpamEvent('b', Point2D()) ])
    traces.append([
        SpamEvent('b', Point2D(x=-1)),
        TimerEvent(),
        SpamEvent('b', Point2D(x=-2)),
    ])
    # invalid
    traces.append([ BehaviourEvent(Point2D(x=1), MonitorState.FALSE) ])
    traces.append([
        SpamEvent('b', Point2D()),
        TimerEvent(),
        BehaviourEvent(Point2D(x=1), MonitorState.FALSE),
        TimerEvent(),
        SpamEvent('b', Point2D(x=1)),
    ])
    return (text, traces)

###############################################################################
# Global Scope With Timeout
###############################################################################

def globally_no_within():
    text = 'globally: no b {x > 0} within 3 s'
    traces = []
    # valid
    traces.append([])
    traces.append([ TimerEvent() ])
    traces.append([ SpamEvent('b', Point2D()) ])
    traces.append([
        SpamEvent('b', Point2D(x=-1)),
        TimerEvent(),
        SpamEvent('b', Point2D(x=-2), state=MonitorState.TRUE),
    ])
    traces.append([
        TimerEvent(),
        TimerEvent(),
        TimerEvent(state=MonitorState.TRUE),
    ])
    traces.append([
        TimerEvent(),
        TimerEvent(),
        SpamEvent('b', Point2D(x=1), state=MonitorState.TRUE),
    ])
    # invalid
    traces.append([ BehaviourEvent(Point2D(x=1), MonitorState.FALSE) ])
    traces.append([
        SpamEvent('b', Point2D()),
        BehaviourEvent(Point2D(x=1), MonitorState.FALSE),
        TimerEvent(),
        SpamEvent('b', Point2D(x=1)),
    ])
    return (text, traces)

###############################################################################
# After Scope No Timeout
###############################################################################

def after_no():
    text = 'after p as P {x > 0}: no b {x > @P.x}'
    traces = []
    # valid
    traces.append([])
    traces.append([ TimerEvent() ])
    traces.append([ SpamEvent('b', Point2D()) ])
    traces.append([ ActivatorEvent(Point2D(x=1), MonitorState.ACTIVE) ])
    traces.append([
        SpamEvent('b', Point2D(x=-1)),
        ActivatorEvent(Point2D(x=1), MonitorState.ACTIVE),
        TimerEvent(),
        SpamEvent('b', Point2D(x=1)),
    ])
    # invalid
    traces.append([
        ActivatorEvent(Point2D(x=1), MonitorState.ACTIVE),
        BehaviourEvent(Point2D(x=2), MonitorState.FALSE)
    ])
    traces.append([
        SpamEvent('b', Point2D()),
        TimerEvent(),
        ActivatorEvent(Point2D(x=1), MonitorState.ACTIVE),
        SpamEvent('b', Point2D(x=1)),
        BehaviourEvent(Point2D(x=2), MonitorState.FALSE),
        TimerEvent(),
        SpamEvent('b', Point2D(x=1)),
    ])
    return (text, traces)

###############################################################################
# After Scope With Timeout
###############################################################################

def after_no_within():
    text = 'after p as P {x > 0}: no b {x > @P.x} within 3 s'
    traces = []
    # valid
    traces.append([])
    traces.append([ TimerEvent() ])
    traces.append([ SpamEvent('b', Point2D()) ])
    traces.append([ ActivatorEvent(Point2D(x=1), MonitorState.ACTIVE) ])
    traces.append([
        SpamEvent('b', Point2D(x=1)),
        ActivatorEvent(Point2D(x=1), MonitorState.ACTIVE),
        TimerEvent(),
        SpamEvent('b', Point2D(x=1)),
        SpamEvent('b', Point2D(x=2), state=MonitorState.TRUE),
        SpamEvent('b', Point2D(x=2)),
    ])
    # invalid
    traces.append([
        ActivatorEvent(Point2D(x=1), MonitorState.ACTIVE),
        BehaviourEvent(Point2D(x=2), MonitorState.FALSE)
    ])
    traces.append([
        SpamEvent('b', Point2D()),
        TimerEvent(),
        ActivatorEvent(Point2D(x=1), MonitorState.ACTIVE),
        SpamEvent('b', Point2D(x=1)),
        BehaviourEvent(Point2D(x=2), MonitorState.FALSE),
        TimerEvent(),
        SpamEvent('b', Point2D(x=1)),
    ])
    return (text, traces)

###############################################################################
# Until Scope No Timeout
###############################################################################

def until_no():
    text = 'until q {x > 0}: no b {x > 0}'
    traces = []
    # valid
    traces.append([])
    traces.append([ TimerEvent() ])
    traces.append([ SpamEvent('b', Point2D()) ])
    traces.append([ TerminatorEvent(Point2D(x=1), MonitorState.TRUE) ])
    traces.append([
        SpamEvent('b', Point2D(x=-1)),
        TimerEvent(),
        SpamEvent('b', Point2D(x=-2)),
    ])
    traces.append([
        SpamEvent('b', Point2D()),
        TimerEvent(),
        TerminatorEvent(Point2D(x=1), MonitorState.TRUE),
        SpamEvent('b', Point2D(x=2)),
    ])
    # invalid
    traces.append([ BehaviourEvent(Point2D(x=1), MonitorState.FALSE) ])
    traces.append([
        SpamEvent('b', Point2D()),
        TimerEvent(),
        BehaviourEvent(Point2D(x=1), MonitorState.FALSE),
        TimerEvent(),
        SpamEvent('b', Point2D(x=1)),
    ])
    traces.append([
        BehaviourEvent(Point2D(x=1), MonitorState.FALSE),
        SpamEvent('q', Point2D(x=1)),
    ])
    return (text, traces)

###############################################################################
# Until Scope With Timeout
###############################################################################

def until_no_within():
    text = 'until q {x > 0}: no b {x > 0} within 3 s'
    traces = []
    # valid
    traces.append([])
    traces.append([ TimerEvent() ])
    traces.append([ SpamEvent('b', Point2D()) ])
    traces.append([ TerminatorEvent(Point2D(x=1), MonitorState.TRUE) ])
    traces.append([
        SpamEvent('b', Point2D()),
        TimerEvent(),
        SpamEvent('b', Point2D(x=1), state=MonitorState.TRUE),
        SpamEvent('q', Point2D(x=1)),
    ])
    traces.append([
        SpamEvent('b', Point2D()),
        TerminatorEvent(Point2D(x=1), MonitorState.TRUE),
        SpamEvent('b', Point2D(x=2)),
    ])
    # invalid
    traces.append([ BehaviourEvent(Point2D(x=1), MonitorState.FALSE) ])
    traces.append([
        SpamEvent('b', Point2D()),
        BehaviourEvent(Point2D(x=1), MonitorState.FALSE),
        TimerEvent(),
        SpamEvent('b', Point2D(x=1)),
    ])
    traces.append([
        BehaviourEvent(Point2D(x=1), MonitorState.FALSE),
        SpamEvent('q', Point2D(x=1)),
    ])
    return (text, traces)

###############################################################################
# After-Until Scope No Timeout
###############################################################################

def after_until_no():
    text = 'after p as P {x + y > 0} until q {x > @P.x}: no b {x > @P.x}'
    traces = []
    # valid
    traces.append([])
    traces.append([ TimerEvent() ])
    traces.append([ SpamEvent('b', Point2D()) ])
    traces.append([ ActivatorEvent(Point2D(x=-2, y=3), MonitorState.ACTIVE) ])
    traces.append([
        SpamEvent('b', Point2D()),
        ActivatorEvent(Point2D(x=1), MonitorState.ACTIVE),
        TimerEvent(),
        SpamEvent('b', Point2D(x=1)),
    ])
    traces.append([
        SpamEvent('p', Point2D()),
        SpamEvent('q', Point2D()),
        SpamEvent('b', Point2D()),
        ActivatorEvent(Point2D(x=1), MonitorState.ACTIVE),
        SpamEvent('b', Point2D(x=1)),
        SpamEvent('q', Point2D(x=1)),
        TerminatorEvent(Point2D(x=2), MonitorState.INACTIVE),
        SpamEvent('b', Point2D(x=2)),
        SpamEvent('q', Point2D(x=2)),
        ActivatorEvent(Point2D(x=2), MonitorState.ACTIVE),
        SpamEvent('b', Point2D(x=2)),
        SpamEvent('q', Point2D(x=2)),
        TerminatorEvent(Point2D(x=3), MonitorState.INACTIVE),
    ])
    # invalid
    traces.append([
        ActivatorEvent(Point2D(x=1), MonitorState.ACTIVE),
        BehaviourEvent(Point2D(x=2), MonitorState.FALSE)
    ])
    traces.append([
        SpamEvent('b', Point2D()),
        TimerEvent(),
        ActivatorEvent(Point2D(x=1), MonitorState.ACTIVE),
        SpamEvent('b', Point2D(x=1)),
        BehaviourEvent(Point2D(x=2), MonitorState.FALSE),
        TimerEvent(),
        SpamEvent('b', Point2D(x=1)),
    ])
    traces.append([
        SpamEvent('p', Point2D()),
        SpamEvent('q', Point2D()),
        SpamEvent('b', Point2D()),
        ActivatorEvent(Point2D(x=1), MonitorState.ACTIVE),
        SpamEvent('b', Point2D(x=1)),
        SpamEvent('q', Point2D(x=1)),
        TerminatorEvent(Point2D(x=2), MonitorState.INACTIVE),
        SpamEvent('b', Point2D(x=2)),
        SpamEvent('q', Point2D(x=2)),
        ActivatorEvent(Point2D(x=2), MonitorState.ACTIVE),
        BehaviourEvent(Point2D(x=3), MonitorState.FALSE),
        SpamEvent('b', Point2D(x=3)),
        SpamEvent('q', Point2D(x=3)),
    ])
    return (text, traces)

###############################################################################
# After-Until Scope With Timeout
###############################################################################

def after_until_no_within():
    text = 'after p as P {x + y > 0} until q {x > @P.x}: no b {x > @P.x} within 3 s'
    traces = []
    # valid
    traces.append([])
    traces.append([ TimerEvent() ])
    traces.append([ SpamEvent('b', Point2D()) ])
    traces.append([ ActivatorEvent(Point2D(x=-2, y=3), MonitorState.ACTIVE) ])
    traces.append([
        SpamEvent('b', Point2D()),
        ActivatorEvent(Point2D(x=1), MonitorState.ACTIVE),
        TimerEvent(),
        SpamEvent('b', Point2D(x=1)),
        TimerEvent(state=MonitorState.SAFE),
        SpamEvent('b', Point2D(x=2)),
    ])
    traces.append([
        SpamEvent('p', Point2D()),
        SpamEvent('q', Point2D(x=1)),
        SpamEvent('b', Point2D(x=1)),
        ActivatorEvent(Point2D(x=1), MonitorState.ACTIVE),
        SpamEvent('b', Point2D(x=1)),
        SpamEvent('q', Point2D(x=1)),
        TerminatorEvent(Point2D(x=2), MonitorState.INACTIVE),
        SpamEvent('b', Point2D(x=2)),
        SpamEvent('q', Point2D(x=2)),
        ActivatorEvent(Point2D(x=2), MonitorState.ACTIVE),
        SpamEvent('b', Point2D(x=2)),
        SpamEvent('q', Point2D(x=2)),
        TimerEvent(state=MonitorState.SAFE),
        TerminatorEvent(Point2D(x=3), MonitorState.INACTIVE),
    ])
    # invalid
    traces.append([
        ActivatorEvent(Point2D(x=1), MonitorState.ACTIVE),
        BehaviourEvent(Point2D(x=2), MonitorState.FALSE)
    ])
    traces.append([
        SpamEvent('b', Point2D()),
        TimerEvent(),
        ActivatorEvent(Point2D(x=1), MonitorState.ACTIVE),
        SpamEvent('b', Point2D(x=1)),
        BehaviourEvent(Point2D(x=2), MonitorState.FALSE),
        TimerEvent(),
        SpamEvent('b', Point2D(x=1)),
    ])
    traces.append([
        SpamEvent('p', Point2D()),
        SpamEvent('q', Point2D()),
        SpamEvent('b', Point2D()),
        ActivatorEvent(Point2D(x=1), MonitorState.ACTIVE),
        SpamEvent('b', Point2D(x=1)),
        SpamEvent('q', Point2D(x=1)),
        TerminatorEvent(Point2D(x=2), MonitorState.INACTIVE),
        SpamEvent('b', Point2D(x=2)),
        SpamEvent('q', Point2D(x=2)),
        ActivatorEvent(Point2D(x=2), MonitorState.ACTIVE),
        BehaviourEvent(Point2D(x=3), MonitorState.FALSE),
        SpamEvent('b', Point2D(x=3)),
        SpamEvent('q', Point2D(x=3)),
    ])
    return (text, traces)



# SPDX-License-Identifier: MIT
# Copyright © 2023 André Santos

"""
Module that contains the 'gen' command line program and the
high-level interface to generate runtime monitors from HPL properties.
"""

###############################################################################
# Imports
###############################################################################

from typing import Any, Dict, Final, Iterable, List, Optional, Type, Union

import argparse
from pathlib import Path

from attrs import field, frozen
from hpl.ast import HplProperty, HplSpecification
from hpl.parser import property_parser, specification_parser
from jinja2 import Environment, PackageLoader

from hplrv.monitors import (
    AbsenceBuilder,
    ExistenceBuilder,
    PreventionBuilder,
    RequirementBuilder,
    ResponseBuilder,
)

###############################################################################
# Constants
###############################################################################

PROG_GEN: Final[str] = 'hpl-rv gen'

ANY_PATH: Final[Type] = Union[Path, str]
ANY_SPEC: Final[Type] = Union[HplSpecification, str]
ANY_PROP: Final[Type] = Union[HplProperty, str]

###############################################################################
# Public Interface
###############################################################################


def monitors_from_files(paths: List[ANY_PATH], lang: str = 'py') -> List[str]:
    """
    Produces a list of monitor code snippets,
    given a list of paths to HPL files with specifications.
    """
    parser = specification_parser()
    r = MonitorGenerator(lang=lang)
    outputs: List[str] = []
    for input_path in paths:
        path: Path = Path(input_path).resolve(strict=True)
        text: str = path.read_text(encoding='utf-8').strip()
        spec: HplSpecification = parser.parse(text)
        for hpl_property in spec.properties:
            outputs.append(r.monitor_class(hpl_property))
    return outputs


def lib_from_files(paths: List[ANY_PATH], lang: str = 'py') -> str:
    """
    Produces a self-contained library of monitors,
    given a list of paths to HPL files with specifications.
    """
    parser = specification_parser()
    r = MonitorGenerator(lang=lang)
    properties: List[HplProperty] = []
    for input_path in paths:
        path: Path = Path(input_path).resolve(strict=True)
        text: str = path.read_text(encoding='utf-8').strip()
        spec: HplSpecification = parser.parse(text)
        properties.extend(spec.properties)
    return r.monitor_library(properties)


def monitors_from_spec(spec: ANY_SPEC, lang: str = 'py') -> List[str]:
    """
    Produces a list of monitor code snippets,
    given an HPL specification.
    """
    if not isinstance(spec, HplSpecification):
        parser = specification_parser()
        spec = parser.parse(spec)
    r = MonitorGenerator(lang=lang)
    outputs: List[str] = []
    for hpl_property in spec.properties:
        outputs.append(r.monitor_class(hpl_property))
    return outputs


def lib_from_spec(spec: ANY_SPEC, lang: str = 'py') -> str:
    """
    Produces a self-contained library of monitors,
    given an HPL specification.
    """
    if not isinstance(spec, HplSpecification):
        parser = specification_parser()
        spec = parser.parse(spec)
    r = MonitorGenerator(lang=lang)
    return r.monitor_library(spec.properties)


def monitor_from_property(property: ANY_PROP, lang: str = 'py') -> str:
    """
    Produces a monitor code snippet, given an HPL property.
    """
    if not isinstance(property, HplProperty):
        parser = property_parser()
        property = parser.parse(property)
    r = MonitorGenerator(lang=lang)
    return r.monitor_class(property)


def monitors_from_properties(properties: List[ANY_PROP], lang: str = 'py') -> List[str]:
    """
    Produces a list of monitor code snippets,
    given a list of HPL properties.
    """
    parser = property_parser()
    r = MonitorGenerator(lang=lang)
    outputs = []
    for property in properties:
        if not isinstance(property, HplProperty):
            property = parser.parse(property)
        outputs.append(r.monitor_class(property))
    return outputs


def lib_from_properties(properties: List[ANY_PROP], lang: str = 'py') -> str:
    """
    Produces a self-contained library of monitors,
    given a list of HPL properties.
    """
    parser = property_parser()
    r = MonitorGenerator(lang=lang)
    properties = [
        parser.parse(property)
        if not isinstance(property, HplProperty)
        else property
        for property in properties
    ]
    return r.monitor_library(properties)


@frozen
class TemplateRenderer:
    jinja_env: Environment

    @classmethod
    def from_pkg_data(
        cls,
        pkg: str = 'hplrv',
        template_dir: str = 'templates'
    ) -> 'TemplateRenderer':
        return cls(Environment(
            loader=PackageLoader(pkg, template_dir),
            line_statement_prefix=None,
            line_comment_prefix=None,
            trim_blocks=True,
            lstrip_blocks=True,
            autoescape=False,
        ))

    def render_template(
        self,
        template_file: str,
        data: Dict[str, Any],
        strip: bool = True,
        encoding: Optional[str] = None
    ) -> str:
        template = self.jinja_env.get_template(template_file)
        text = template.render(**data)
        if strip:
            text = text.strip()
        if encoding is None:
            return text
        return text.encode(encoding)


@frozen
class MonitorGenerator:
    renderer: TemplateRenderer = field(factory=TemplateRenderer.from_pkg_data)
    lang: str = 'py'

    def monitor_library(
        self,
        spec_or_properties: Union[Iterable[HplProperty], HplSpecification],
    ) -> str:
        data = self.data_for_monitor_library(spec_or_properties)
        template_file = f'{self.lang}/library.{self.lang}.jinja'
        return self.renderer.render_template(template_file, data)

    def data_for_monitor_library(
        self,
        spec_or_properties: Union[Iterable[HplProperty], HplSpecification],
    ) -> Dict[str, Any]:
        class_names = []
        callbacks = {}
        monitor_classes = []
        if isinstance(spec_or_properties, HplSpecification):
            spec_or_properties = spec_or_properties.properties
        for p in spec_or_properties:
            builder, template_file = self._template(p, True)
            i = len(class_names)
            builder.class_name = f'Property{i}Monitor'
            class_names.append(builder.class_name)
            for name in builder.on_msg:
                if name not in callbacks:
                    callbacks[name] = set()
                callbacks[name].add(i)
            data = {'state_machine': builder}
            monitor_classes.append(self.renderer.render_template(template_file, data))
        return {
            'class_names': class_names,
            'monitor_classes': monitor_classes,
            'callbacks': callbacks,
        }

    def monitor_class(
        self,
        hpl_property: HplProperty,
        class_name: Optional[str] = None,
        id_as_class: bool = True,
        encoding: Optional[str] = None,
    ) -> str:
        data = self.data_for_monitor_class(
            hpl_property,
            class_name=class_name,
            id_as_class=id_as_class,
        )
        template_file = data['template_file']
        return self.renderer.render_template(template_file, data, encoding=encoding)

    def data_for_monitor_class(
        self,
        hpl_property: HplProperty,
        class_name: Optional[str] = None,
        id_as_class: bool = True,
    ) -> Dict[str, Any]:
        builder, template_file = self._template(hpl_property, id_as_class)
        if class_name:
            builder.class_name = class_name
        return {
            'template_file': template_file,
            'state_machine': builder,
        }

    def _template(self, hpl_property, id_as_class):
        if hpl_property.pattern.is_absence:
            builder = AbsenceBuilder(hpl_property)
            template_file = f'{self.lang}/absence.{self.lang}.jinja'
        elif hpl_property.pattern.is_existence:
            builder = ExistenceBuilder(hpl_property)
            template_file = f'{self.lang}/existence.{self.lang}.jinja'
        elif hpl_property.pattern.is_requirement:
            builder = RequirementBuilder(hpl_property)
            if not builder.has_trigger_refs:
                template_file = f'{self.lang}/requirement-simple.{self.lang}.jinja'
            else:
                template_file = f'{self.lang}/requirement-refs.{self.lang}.jinja'
        elif hpl_property.pattern.is_response:
            builder = ResponseBuilder(hpl_property)
            template_file = f'{self.lang}/response.{self.lang}.jinja'
        elif hpl_property.pattern.is_prevention:
            builder = PreventionBuilder(hpl_property)
            template_file = f'{self.lang}/prevention.{self.lang}.jinja'
        else:
            raise ValueError('unknown pattern: ' + str(hpl_property.pattern))
        if id_as_class:
            name = hpl_property.metadata.get('id', 'Property')
            name = ''.join(word.title() for word in name.split("_") if word)
            builder.class_name = name + 'Monitor'
        return (builder, template_file)


###############################################################################
# Entry Point
###############################################################################


def subprogram(
    argv: Optional[List[str]],
    _settings: Optional[Dict[str, Any]] = None,
) -> int:
    args = parse_arguments(argv)
    return run(args, _settings or {})


def run(args: Dict[str, Any], _settings: Dict[str, Any]) -> int:
    parts: List[str] = []
    lang: str = args['lang']
    if args.get('files'):
        if args.get('just_classes'):
            parts.extend(monitors_from_files(args['args'], lang=lang))
        else:
            parts.append(lib_from_files(args['args'], lang=lang))
    else:
        if args.get('just_classes'):
            parts.extend(monitors_from_properties(args['args'], lang=lang))
        else:
            parts.append(lib_from_properties(args['args'], lang=lang))
    output: str = '\n\n'.join(code for code in parts)

    input_path: str = args.get('output')
    if input_path:
        path: Path = Path(input_path).resolve(strict=False)
        path.write_text(output, encoding='utf-8')
    else:
        print(output)
    return 0


###############################################################################
# Argument Parsing
###############################################################################


def parse_arguments(argv: Optional[List[str]]) -> Dict[str, Any]:
    description = 'Generate runtime monitors from HPL properties.'
    parser = argparse.ArgumentParser(prog=PROG_GEN, description=description)

    parser.add_argument('-o', '--output', help='output file to place generated code')

    parser.add_argument(
        '-f',
        '--files',
        action='store_true',
        help='process args as HPL files (default: HPL properties)',
    )

    parser.add_argument(
        '-c',
        '--class',
        dest='just_classes',
        action='store_true',
        help='output just the monitor classes (default: false)',
    )

    parser.add_argument(
        '-l',
        '--lang',
        choices=('py', 'js'),
        default='py',
        help='language of the generated code (default: py)',
    )

    parser.add_argument('args', nargs='+', help='input properties')

    args = parser.parse_args(args=argv)
    return vars(args)



# SPDX-License-Identifier: MIT
# Copyright © 2023 André Santos

###############################################################################
# Imports
###############################################################################

import asyncio
import importlib.util
from pathlib import Path
import sys
from threading import Thread

from hplrv.gen import lib_from_properties
from ..common_data import Point2D

###############################################################################
# Test Setup
###############################################################################

MODULE_NAME = 'generated_monitors'
FILE_NAME = f'{MODULE_NAME}.py'
FILE_PATH = Path(__file__).resolve(strict=True).parent / FILE_NAME
HOST = '127.0.0.1'
PORT = 4242


def example_properties():
    return [
        'globally: no /a {x < 0}',
        'globally: no /a {x > 0} within 100 ms',
    ]


def generate_monitors():
    code: str = lib_from_properties(example_properties())
    FILE_PATH.write_text(code, encoding='utf8')


def import_generated_monitors():
    spec = importlib.util.spec_from_file_location(MODULE_NAME, FILE_PATH)
    module = importlib.util.module_from_spec(spec)
    sys.modules[MODULE_NAME] = module
    spec.loader.exec_module(module)
    return module


###############################################################################
# Integration Test
###############################################################################


def test_live_monitoring_server():
    generate_monitors()
    lib = import_generated_monitors()
    man = lib.HplMonitorManager()
    man.live_server.host = HOST
    man.live_server.port = PORT
    thread: Thread = man.live_server.start_thread()
    asyncio.run(integration_task(man, thread))


async def integration_task(man, thread: Thread):
    reader, writer = await asyncio.open_connection(HOST, PORT)
    report = await reader.readline()
    assert report.decode('utf8').startswith('[')
    man.launch(0.0)
    man.on_timer(0.2)
    assert man.monitors[1].verdict is True
    report = await reader.readline()
    assert report.decode('utf8').startswith('{')
    man.on_msg__a(Point2D(x=-1), 0.5)
    assert man.monitors[0].verdict is False
    report = await reader.readline()
    assert report.decode('utf8').startswith('{')
    man.shutdown(1.0)
    thread.join(10.0)
    assert not thread.is_alive()
